# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import sys
from importlib import reload
from types import ModuleType

import pytest

from test.widgets.conftest import FakeBar


def no_op(*args, **kwargs):
    pass


async def mock_signal_receiver(*args, **kwargs):
    return True


def fake_timer(interval, func, *args, **kwargs):
    class TimerObj:
        def cancel(self):
            pass

        @property
        def _scheduled(self):
            return False

    return TimerObj()


class MockConstants(ModuleType):
    class MessageType:
        SIGNAL = 1


class MockMessage:
    def __init__(self, is_signal=True, body=None):
        self.message_type = 1 if is_signal else 0
        self.body = body


# dbus_next message data is stored in variants. The widget extracts the
# information via the `value` attribute so we just need to mock that here.
class obj:  # noqa: N801
    def __init__(self, value):
        self.value = value


# Creates a mock message body containing both metadata and playback status
def metadata_and_status(status):
    return MockMessage(
        body=(
            "",
            {
                "Metadata": obj(
                    {
                        "mpris:trackid": obj(1),
                        "xesam:url": obj("/path/to/rickroll.mp3"),
                        "xesam:title": obj("Never Gonna Give You Up"),
                        "xesam:artist": obj(["Rick Astley"]),
                        "xesam:album": obj("Whenever You Need Somebody"),
                        "mpris:length": obj(200000000),
                    }
                ),
                "PlaybackStatus": obj(status),
            },
            [],
        )
    )


# Creates a mock message body containing just playback status
def playback_status(status, signal=True):
    return MockMessage(is_signal=signal, body=("", {"PlaybackStatus": obj(status)}, []))


METADATA_PLAYING = metadata_and_status("Playing")
METADATA_PAUSED = metadata_and_status("Paused")
STATUS_PLAYING = playback_status("Playing")
STATUS_PAUSED = playback_status("Paused")
STATUS_STOPPED = playback_status("Stopped")
NON_SIGNAL = playback_status("Paused", False)


@pytest.fixture
def patched_module(monkeypatch):
    # Remove dbus_next.constants entry from modules. If it's not there, don't raise error
    monkeypatch.delitem(sys.modules, "dbus_next.constants", raising=False)
    monkeypatch.setitem(sys.modules, "dbus_next.constants", MockConstants("dbus_next.constants"))
    from libqtile.widget import mpris2widget

    # Need to force reload of the module to ensure patched module is loaded
    # This may only be needed if dbus_next is installed on testing system so helpful for
    # local tests.
    reload(mpris2widget)
    monkeypatch.setattr("libqtile.widget.mpris2widget.add_signal_receiver", mock_signal_receiver)
    return mpris2widget


def test_mpris2_signal_handling(fake_qtile, patched_module, fake_window):
    mp = patched_module.Mpris2()
    fakebar = FakeBar([mp], window=fake_window)
    mp.timeout_add = fake_timer
    mp._configure(fake_qtile, fakebar)

    assert mp.displaytext == ""

    # No text will be displayed if widget is not configured
    mp.parse_message(*METADATA_PLAYING.body)
    assert mp.displaytext == ""

    # Set configured flag, create a message with the metadata and playback status
    mp.configured = True
    mp.parse_message(*METADATA_PLAYING.body)
    assert mp.text == "Never Gonna Give You Up - Whenever You Need Somebody - Rick Astley"

    # If widget receives "paused" signal it prefixes track with "Paused: "
    mp.parse_message(*STATUS_PAUSED.body)
    assert mp.text == "Paused: Never Gonna Give You Up - Whenever You Need Somebody - Rick Astley"

    # If widget receives "stopped" signal with no metadata then widget is blank
    mp.parse_message(*STATUS_STOPPED.body)
    assert mp.displaytext == ""

    # Reset to playing + metadata
    mp.parse_message(*METADATA_PLAYING.body)
    assert mp.text == "Never Gonna Give You Up - Whenever You Need Somebody - Rick Astley"

    # If widget receives "paused" signal with metadata then message is "Paused: {metadata}"
    mp.parse_message(*METADATA_PAUSED.body)
    assert mp.text == "Paused: Never Gonna Give You Up - Whenever You Need Somebody - Rick Astley"

    # If widget now receives "playing" signal with no metadata, "paused" word is removed
    mp.parse_message(*STATUS_PLAYING.body)
    assert mp.text == "Never Gonna Give You Up - Whenever You Need Somebody - Rick Astley"

    info = mp.info()
    assert info["text"] == "Never Gonna Give You Up - Whenever You Need Somebody - Rick Astley"
    assert info["isplaying"]


def test_mpris2_custom_stop_text(fake_qtile, patched_module, fake_window):
    mp = patched_module.Mpris2(stop_pause_text="Test Paused")
    fakebar = FakeBar([mp], window=fake_window)
    mp.timeout_add = fake_timer
    mp._configure(fake_qtile, fakebar)
    mp.configured = True

    mp.parse_message(*METADATA_PLAYING.body)
    assert mp.text == "Never Gonna Give You Up - Whenever You Need Somebody - Rick Astley"

    # Check our custom paused wording is shown
    mp.parse_message(*STATUS_PAUSED.body)
    assert mp.text == "Test Paused"


def test_mpris2_no_metadata(fake_qtile, patched_module, fake_window):
    mp = patched_module.Mpris2()
    fakebar = FakeBar([mp], window=fake_window)
    mp.timeout_add = fake_timer
    mp._configure(fake_qtile, fakebar)
    mp.configured = True

    mp.parse_message(*STATUS_PLAYING.body)
    assert mp.text == "No metadata for current track"


def test_mpris2_no_scroll(fake_qtile, patched_module, fake_window):
    # If no scrolling, then the update function creates the text to display
    # and draws the bar.
    mp = patched_module.Mpris2(scroll_chars=None)
    fakebar = FakeBar([mp], window=fake_window)
    mp.timeout_add = fake_timer
    mp._configure(fake_qtile, fakebar)
    mp.configured = True

    mp.parse_message(*METADATA_PLAYING.body)
    assert mp.text == "Never Gonna Give You Up - Whenever You Need Somebody - Rick Astley"

    mp.parse_message(*METADATA_PAUSED.body)
    assert mp.text == "Paused: Never Gonna Give You Up - Whenever You Need Somebody - Rick Astley"


def test_mpris2_deprecated_format(patched_module):
    """
    Previously, metadata was displayed by using a list of fields.
    Now, we use a `format` string. The widget should create this when a user
    provides `display_metadata` in their config.
    """
    mp = patched_module.Mpris2(display_metadata=["xesam:title", "xesam:artist"])
    assert mp.format == "{xesam:title} - {xesam:artist}"



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

import sys
from importlib import reload
from types import ModuleType

import pytest

from libqtile.widget import generic_poll_text


class Mockxml(ModuleType):
    @classmethod
    def parse(cls, value):
        return {"test": value}


class MockRequest:
    return_value = None

    def __init__(self, *args, **kwargs):
        pass


class Mockurlopen:
    def __init__(self, request):
        self.request = request

    class headers:  # noqa: N801
        @classmethod
        def get_content_charset(cls):
            return "utf-8"

    def read(self):
        return self.request.return_value


def test_gen_poll_text():
    gpt_no_func = generic_poll_text.GenPollText()
    assert gpt_no_func.poll() == "You need a poll function"

    gpt_with_func = generic_poll_text.GenPollText(func=lambda: "Has function")
    assert gpt_with_func.poll() == "Has function"


def test_gen_poll_url_not_configured():
    gpurl = generic_poll_text.GenPollUrl()
    assert gpurl.poll() == "Invalid config"


def test_gen_poll_url_no_json():
    gpurl = generic_poll_text.GenPollUrl(json=False)
    assert "Content-Type" not in gpurl.headers


def test_gen_poll_url_headers_and_json():
    gpurl = generic_poll_text.GenPollUrl(
        headers={"fake-header": "fake-value"},
        data={"argument": "data value"},
        user_agent="qtile test",
    )

    assert gpurl.headers["User-agent"] == "qtile test"
    assert gpurl.headers["fake-header"] == "fake-value"
    assert gpurl.headers["Content-Type"] == "application/json"
    assert gpurl.data.decode() == '{"argument": "data value"}'


def test_gen_poll_url_text(monkeypatch):
    gpurl = generic_poll_text.GenPollUrl(json=False, parse=lambda x: x, url="testing")
    monkeypatch.setattr(generic_poll_text, "Request", MockRequest)
    monkeypatch.setattr(generic_poll_text, "urlopen", Mockurlopen)
    generic_poll_text.Request.return_value = b"OK"
    assert gpurl.poll() == "OK"


def test_gen_poll_url_json(monkeypatch):
    gpurl = generic_poll_text.GenPollUrl(parse=lambda x: x, data=[1, 2, 3], url="testing")
    monkeypatch.setattr(generic_poll_text, "Request", MockRequest)
    monkeypatch.setattr(generic_poll_text, "urlopen", Mockurlopen)
    generic_poll_text.Request.return_value = b'{"test": "OK"}'
    assert gpurl.poll()["test"] == "OK"


def test_gen_poll_url_xml_no_xmltodict(monkeypatch):
    gpurl = generic_poll_text.GenPollUrl(json=False, xml=True, parse=lambda x: x, url="testing")
    monkeypatch.setattr(generic_poll_text, "Request", MockRequest)
    monkeypatch.setattr(generic_poll_text, "urlopen", Mockurlopen)
    generic_poll_text.Request.return_value = b"OK"
    with pytest.raises(Exception):
        gpurl.poll()


def test_gen_poll_url_xml_has_xmltodict(monkeypatch):
    # injected fake xmltodict module but we have to reload the widget module
    # as the ImportError test is only run once when the module is loaded.
    monkeypatch.setitem(sys.modules, "xmltodict", Mockxml("xmltodict"))
    reload(generic_poll_text)
    gpurl = generic_poll_text.GenPollUrl(json=False, xml=True, parse=lambda x: x, url="testing")
    monkeypatch.setattr(generic_poll_text, "Request", MockRequest)
    monkeypatch.setattr(generic_poll_text, "urlopen", Mockurlopen)
    generic_poll_text.Request.return_value = b"OK"
    assert gpurl.poll()["test"] == "OK"


def test_gen_poll_url_broken_parse(monkeypatch):
    gpurl = generic_poll_text.GenPollUrl(json=False, parse=lambda x: x.foo, url="testing")
    monkeypatch.setattr(generic_poll_text, "Request", MockRequest)
    monkeypatch.setattr(generic_poll_text, "urlopen", Mockurlopen)
    generic_poll_text.Request.return_value = b"OK"
    assert gpurl.poll() == "Can't parse"



# Copyright (c) 2017 Dario Giovannetti
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

import libqtile
import libqtile.config
import libqtile.hook
from libqtile import layout
from libqtile.confreader import Config
from test.layouts.layout_utils import (
    assert_dimensions_fit,
    assert_focus_path_unordered,
    assert_focused,
)

try:
    # Check to see if we should skip tests using notifications on Wayland
    import gi

    gi.require_version("Gtk", "3.0")
    gi.require_version("GtkLayerShell", "0.1")
    from gi.repository import GtkLayerShell  # noqa: F401

    has_wayland_notifications = True
except (ImportError, ValueError):
    has_wayland_notifications = False


class AllLayoutsConfig(Config):
    """
    Ensure that all layouts behave consistently in some common scenarios.
    """

    groups = [
        libqtile.config.Group("a"),
        libqtile.config.Group("b"),
        libqtile.config.Group("c"),
        libqtile.config.Group("d"),
    ]
    follow_mouse_focus = False
    floating_layout = libqtile.resources.default_config.floating_layout
    screens = []

    @staticmethod
    def iter_layouts():
        # Retrieve the layouts dynamically (i.e. do not hard-code a list) to
        # prevent forgetting to add new future layouts
        for layout_name in dir(layout):
            layout_cls = getattr(layout, layout_name)
            try:
                test = issubclass(layout_cls, layout.base.Layout)
            except TypeError:
                pass
            else:
                # Explicitly exclude the Slice layout, since it depends on
                # other layouts (tested here) and has its own specific tests
                if test and layout_name != "Slice":
                    yield layout_name, layout_cls

    @classmethod
    def generate(cls):
        """
        Generate a configuration for each layout currently in the repo.
        Each configuration has only the tested layout (i.e. 1 item) in the
        'layouts' variable.
        """
        return [
            type(layout_name, (cls,), {"layouts": [layout_cls()]})
            for layout_name, layout_cls in cls.iter_layouts()
        ]


class AllDelegateLayoutsConfig(AllLayoutsConfig):
    @classmethod
    def generate(cls):
        """
        Generate a Slice configuration for each layout currently in the repo.
        Each layout is made a delegate/fallback layout of the Slice layout.
        Each configuration has only the tested layout (i.e. 1 item) in the
        'layouts' variable.
        """
        return [
            type(
                layout_name,
                (cls,),
                {"layouts": [layout.slice.Slice(wname="nevermatch", fallback=layout_cls())]},
            )
            for layout_name, layout_cls in cls.iter_layouts()
        ]


class AllLayouts(AllLayoutsConfig):
    """
    Like AllLayoutsConfig, but all the layouts in the repo are installed
    together in the 'layouts' variable.
    """

    layouts = [layout_cls() for layout_name, layout_cls in AllLayoutsConfig.iter_layouts()]


class AllLayoutsConfigEvents(AllLayoutsConfig):
    """
    Extends AllLayoutsConfig to test events.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # TODO: Test more events

        @libqtile.hook.subscribe.startup
        def _():
            libqtile.qtile.test_data = {
                "focus_change": 0,
                "config_name": self.__class__.__name__,
            }

        @libqtile.hook.subscribe.focus_change
        def _():
            if hasattr(libqtile.qtile, "test_data"):
                # focus_change can fire before during startup
                libqtile.qtile.test_data["focus_change"] += 1


each_layout_config = pytest.mark.parametrize(
    "manager", AllLayoutsConfig.generate(), indirect=True
)
all_layouts_config = pytest.mark.parametrize("manager", [AllLayouts], indirect=True)
each_layout_config_events = pytest.mark.parametrize(
    "manager", AllLayoutsConfigEvents.generate(), indirect=True
)
each_delegate_layout_config = pytest.mark.parametrize(
    "manager", AllDelegateLayoutsConfig.generate(), indirect=True
)


@each_layout_config
def test_window_order_fullscreen(manager):
    # Add a window to fullscreen
    manager.test_window("tofullscreen")

    # windows to add after the fullscreen window
    windows_after = ["two", "three"]

    # Add some windows before
    for win in windows_after:
        manager.test_window(win)

    windows_order = manager.c.layout.info()["clients"]

    # Focus window and toggle fullscreen
    manager.c.group.focus_by_name("tofullscreen")
    manager.c.window.toggle_fullscreen()
    manager.c.window.toggle_fullscreen()

    # Windows must be sorted in the same order as they were created
    assert windows_order == manager.c.layout.info()["clients"]


@each_layout_config
def test_window_types(manager):
    if manager.backend.name == "wayland" and not has_wayland_notifications:
        pytest.skip("Notification tests for Wayland need gtk-layer-shell")

    manager.test_window("one")

    # A dialog should take focus and be floating
    # A notification shouldn't steal focus and should be floating
    manager.test_window("dialog", floating=True)
    assert_focused(manager, "dialog")
    manager.test_notification("notification")
    assert manager.c.group.info()["focus"] == "dialog"
    for window in manager.c.windows():
        if window["name"] in ("dialog", "notification"):
            assert window["floating"]


@each_layout_config
def test_focus_cycle(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("float1", floating=True)
    manager.test_window("float2", floating=True)
    manager.test_window("three")

    # Test preconditions (the order of items in 'clients' is managed by each layout)
    assert set(manager.c.layout.info()["clients"]) == {"one", "two", "three"}
    assert_focused(manager, "three")

    # Assert that the layout cycles the focus on all windows
    assert_focus_path_unordered(manager, "float1", "float2", "one", "two", "three")


@each_layout_config
def test_swap_window_order(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")

    window_order = manager.c.group.info().get("windows")
    assert window_order == ["one", "two", "three"]

    # Swap window on index 0 with index 2
    manager.c.group.focus_by_name("one")
    manager.c.group.swap_window_order(2)
    window_order = manager.c.group.info().get("windows")
    assert window_order == ["three", "two", "one"]
    assert_focused(manager, "one")

    # Swap window on index 1 with index 0
    manager.c.group.focus_by_name("two")
    manager.c.group.swap_window_order(0)
    window_order = manager.c.group.info().get("windows")
    assert window_order == ["two", "three", "one"]
    assert_focused(manager, "two")

    # Swap window on index 0 with index out of bounds
    manager.c.group.focus_by_name("two")
    manager.c.group.swap_window_order(3)
    window_order = manager.c.group.info().get("windows")
    assert window_order == ["two", "three", "one"]
    assert_focused(manager, "two")


@each_layout_config
def test_focus_back(manager):
    # No exception must be raised without windows
    manager.c.group.focus_back()

    # Nothing must happen with only one window
    manager.test_window("one")
    manager.c.group.focus_back()
    assert_focused(manager, "one")

    # 2 windows
    two = manager.test_window("two")
    assert_focused(manager, "two")
    manager.c.group.focus_back()
    assert_focused(manager, "one")
    manager.c.group.focus_back()
    assert_focused(manager, "two")

    # Float a window
    three = manager.test_window("three")
    manager.c.group.focus_back()
    assert_focused(manager, "two")
    manager.c.window.toggle_floating()
    manager.c.group.focus_back()
    assert_focused(manager, "three")

    # If the previous window is killed, the further previous one must be focused
    manager.test_window("four")
    manager.kill_window(two)
    manager.kill_window(three)
    assert_focused(manager, "four")
    manager.c.group.focus_back()
    assert_focused(manager, "one")


# TODO: Test more events
@each_layout_config_events
def test_focus_change_event(manager):
    # Test that the correct number of focus_change events are fired e.g. when
    # opening, closing or switching windows.
    # If for example a layout explicitly fired a focus_change event even though
    # group._Group.focus() or group._Group.remove() already fire one, the other
    # installed layouts would wrongly react to it and cause misbehaviour.
    # In short, this test prevents layouts from influencing each other in
    # unexpected ways.

    # Ensure we are in fact using the right layout
    assert manager.c.get_test_data()["config_name"].lower() == manager.c.layout.info()["name"]

    # Spawning a window must fire only 1 focus_change event
    assert manager.c.get_test_data()["focus_change"] == 0
    one = manager.test_window("one")
    assert manager.c.get_test_data()["focus_change"] == 1
    two = manager.test_window("two")
    assert manager.c.get_test_data()["focus_change"] == 2
    three = manager.test_window("three")
    assert manager.c.get_test_data()["focus_change"] == 3

    # Switching window must fire only 1 focus_change event
    assert_focused(manager, "three")
    manager.c.group.focus_by_name("one")
    assert manager.c.get_test_data()["focus_change"] == 4
    assert_focused(manager, "one")

    # Focusing the current window must fire another focus_change event
    manager.c.group.focus_by_name("one")
    assert manager.c.get_test_data()["focus_change"] == 5

    # Toggling a window floating should not fire focus_change events
    manager.c.window.toggle_floating()
    assert manager.c.get_test_data()["focus_change"] == 5
    manager.c.window.toggle_floating()
    assert manager.c.get_test_data()["focus_change"] == 5

    # Removing the focused window must fire only 1 focus_change event
    assert_focused(manager, "one")
    assert manager.c.group.info()["focus_history"] == ["two", "three", "one"]
    manager.kill_window(one)
    assert manager.c.get_test_data()["focus_change"] == 6

    # The position where 'one' was after it was floated and unfloated
    # above depends on the layout, so we can't predict here what window gets
    # selected after killing it; for this reason, focus 'three' explicitly to
    # continue testing
    manager.c.group.focus_by_name("three")
    assert manager.c.group.info()["focus_history"] == ["two", "three"]
    assert manager.c.get_test_data()["focus_change"] == 7

    # Removing a non-focused window must not fire focus_change events
    manager.kill_window(two)
    assert manager.c.get_test_data()["focus_change"] == 7
    assert_focused(manager, "three")

    # Removing the last window must still generate 1 focus_change event
    manager.kill_window(three)
    assert manager.c.layout.info()["clients"] == []
    assert manager.c.get_test_data()["focus_change"] == 8


@each_layout_config
def test_remove(manager):
    one = manager.test_window("one")
    two = manager.test_window("two")
    three = manager.test_window("three")
    assert_focused(manager, "three")
    assert manager.c.group.info()["focus_history"] == ["one", "two", "three"]

    # Removing a focused window must focus another (which one depends on the layout)
    manager.kill_window(three)
    assert manager.c.window.info()["name"] in manager.c.layout.info()["clients"]

    # To continue testing, explicitly set focus on 'two'
    manager.c.group.focus_by_name("two")
    manager.test_window("four")
    assert_focused(manager, "four")
    assert manager.c.group.info()["focus_history"] == ["one", "two", "four"]

    # Removing a non-focused window must not change the current focus
    manager.kill_window(two)
    assert_focused(manager, "four")
    assert manager.c.group.info()["focus_history"] == ["one", "four"]

    # Add more windows and shuffle the focus order
    five = manager.test_window("five")
    manager.test_window("six")
    manager.c.group.focus_by_name("one")
    seven = manager.test_window("seven")
    manager.c.group.focus_by_name("six")
    assert_focused(manager, "six")
    assert manager.c.group.info()["focus_history"] == ["four", "five", "one", "seven", "six"]

    manager.kill_window(five)
    manager.kill_window(one)
    assert_focused(manager, "six")
    assert manager.c.group.info()["focus_history"] == ["four", "seven", "six"]

    manager.c.group.focus_by_name("seven")
    manager.kill_window(seven)
    assert manager.c.window.info()["name"] in manager.c.layout.info()["clients"]


@each_layout_config
def test_remove_floating(manager):
    one = manager.test_window("one")
    manager.test_window("two")
    float1 = manager.test_window("float1", floating=True)
    assert_focused(manager, "float1")
    assert set(manager.c.layout.info()["clients"]) == {"one", "two"}
    assert manager.c.group.info()["focus_history"] == ["one", "two", "float1"]

    # Removing a focused floating window must focus the one that was focused before
    manager.kill_window(float1)
    assert_focused(manager, "two")
    assert manager.c.group.info()["focus_history"] == ["one", "two"]

    float2 = manager.test_window("float2", floating=True)
    assert_focused(manager, "float2")
    assert manager.c.group.info()["focus_history"] == ["one", "two", "float2"]

    # Removing a non-focused floating window must not change the current focus
    manager.c.group.focus_by_name("two")
    manager.kill_window(float2)
    assert_focused(manager, "two")
    assert manager.c.group.info()["focus_history"] == ["one", "two"]

    # Add more windows and shuffle the focus order
    manager.test_window("three")
    float3 = manager.test_window("float3", floating=True)
    manager.c.group.focus_by_name("one")
    float4 = manager.test_window("float4", floating=True)
    float5 = manager.test_window("float5", floating=True)
    manager.c.group.focus_by_name("three")
    manager.c.group.focus_by_name("float3")
    assert manager.c.group.info()["focus_history"] == [
        "two",
        "one",
        "float4",
        "float5",
        "three",
        "float3",
    ]

    manager.kill_window(one)
    assert_focused(manager, "float3")
    assert manager.c.group.info()["focus_history"] == [
        "two",
        "float4",
        "float5",
        "three",
        "float3",
    ]

    manager.kill_window(float5)
    assert_focused(manager, "float3")
    assert manager.c.group.info()["focus_history"] == ["two", "float4", "three", "float3"]

    # The focus must be given to the previous window even if it's floating
    manager.c.group.focus_by_name("float4")
    assert manager.c.group.info()["focus_history"] == ["two", "three", "float3", "float4"]
    manager.kill_window(float4)
    assert_focused(manager, "float3")
    assert manager.c.group.info()["focus_history"] == ["two", "three", "float3"]

    four = manager.test_window("four")
    float6 = manager.test_window("float6", floating=True)
    five = manager.test_window("five")
    manager.c.group.focus_by_name("float3")
    assert manager.c.group.info()["focus_history"] == [
        "two",
        "three",
        "four",
        "float6",
        "five",
        "float3",
    ]

    # Killing several unfocused windows before the current one, and then
    # killing the current window, must focus the remaining most recently
    # focused window
    manager.kill_window(five)
    manager.kill_window(four)
    manager.kill_window(float6)
    assert manager.c.group.info()["focus_history"] == ["two", "three", "float3"]
    manager.kill_window(float3)
    assert_focused(manager, "three")
    assert manager.c.group.info()["focus_history"] == ["two", "three"]


@each_layout_config
def test_desktop_notifications(manager):
    # Unlike normal floating windows such as dialogs, notifications don't steal
    # focus when they spawn, so test them separately
    if manager.backend.name == "wayland" and not has_wayland_notifications:
        pytest.skip("Notification tests for Wayland need gtk-layer-shell")

    # A notification fired in an empty group must not take focus
    notif1 = manager.test_notification("notif1")
    assert manager.c.group.info()["focus"] is None
    manager.kill_window(notif1)

    # A window is spawned while a notification is displayed
    notif2 = manager.test_notification("notif2")
    one = manager.test_window("one")
    assert manager.c.group.info()["focus_history"] == ["one"]
    manager.kill_window(notif2)

    # Another notification is fired, but the focus must not change
    notif3 = manager.test_notification("notif3")
    assert_focused(manager, "one")
    manager.kill_window(notif3)

    # Complicate the scenario with multiple windows and notifications

    dialog1 = manager.test_window("dialog1", floating=True)
    manager.test_window("two")
    notif4 = manager.test_notification("notif4")
    notif5 = manager.test_notification("notif5")
    assert manager.c.group.info()["focus_history"] == ["one", "dialog1", "two"]

    dialog2 = manager.test_window("dialog2", floating=True)
    manager.kill_window(notif5)
    manager.test_window("three")
    manager.kill_window(one)
    manager.c.group.focus_by_name("two")
    notif6 = manager.test_notification("notif6")
    notif7 = manager.test_notification("notif7")
    manager.kill_window(notif4)
    notif8 = manager.test_notification("notif8")
    assert manager.c.group.info()["focus_history"] == ["dialog1", "dialog2", "three", "two"]

    manager.test_window("dialog3", floating=True)
    manager.kill_window(dialog1)
    manager.kill_window(dialog2)
    manager.kill_window(notif6)
    manager.c.group.focus_by_name("three")
    manager.kill_window(notif7)
    manager.kill_window(notif8)
    assert manager.c.group.info()["focus_history"] == ["two", "dialog3", "three"]


@each_delegate_layout_config
def test_only_uses_delegated_screen_rect(manager):
    manager.test_window("one")
    manager.c.group.focus_by_name("one")
    assert_focused(manager, "one")
    assert_dimensions_fit(manager, 256, 0, 800 - 256, 600)


@all_layouts_config
def test_cycle_layouts(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")
    manager.test_window("four")
    manager.c.group.focus_by_name("three")
    assert_focused(manager, "three")

    # Cycling all the layouts must keep the current window focused
    initial_layout_name = manager.c.layout.info()["name"]
    while True:
        manager.c.next_layout()
        if manager.c.layout.info()["name"] == initial_layout_name:
            break
        # Use manager.c.layout.info()['name'] in the assertion message, so we
        # know which layout is buggy
        assert manager.c.window.info()["name"] == "three", manager.c.layout.info()["name"]

    # Now try backwards
    while True:
        manager.c.prev_layout()
        if manager.c.layout.info()["name"] == initial_layout_name:
            break
        # Use manager.c.layout.info()['name'] in the assertion message, so we
        # know which layout is buggy
        assert manager.c.window.info()["name"] == "three", manager.c.layout.info()["name"]


class AllLayoutsMultipleBorders(AllLayoutsConfig):
    """
    Like AllLayouts, but all the layouts have border_focus set to a list of colors.
    """

    layouts = [
        layout_cls(border_focus=["#000", "#111", "#222", "#333", "#444"])
        for layout_name, layout_cls in AllLayoutsConfig.iter_layouts()
    ]


@pytest.mark.parametrize("manager", [AllLayoutsMultipleBorders], indirect=True)
def test_multiple_borders(manager):
    manager.test_window("one")
    manager.test_window("two")

    initial_layout_name = manager.c.layout.info()["name"]
    while True:
        manager.c.next_layout()
        if manager.c.layout.info()["name"] == initial_layout_name:
            break



# Copyright (c) 2022 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import pytest

import libqtile.config
from libqtile import bar, layout, widget
from libqtile.confreader import Config


class ScreenSplitConfig(Config):
    auto_fullscreen = True
    groups = [libqtile.config.Group("a")]
    layouts = [layout.Max(), layout.ScreenSplit()]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = [
        libqtile.config.Screen(
            top=bar.Bar(
                [widget.ScreenSplit(), widget.ScreenSplit(format="{layout} - {split_name}")], 40
            )
        )
    ]
    follow_mouse_focus = False


screensplit_config = pytest.mark.parametrize("manager", [ScreenSplitConfig], indirect=True)


@screensplit_config
def test_screensplit_text(manager):
    widget = manager.c.widget["screensplit"]
    assert widget.info()["text"] == ""

    manager.c.next_layout()
    assert widget.info()["text"] == "top (max)"

    manager.c.layout.next_split()
    assert widget.info()["text"] == "bottom (columns)"

    manager.c.next_layout()
    assert widget.info()["text"] == ""


@screensplit_config
def test_screensplit_scroll_actions(manager):
    widget = manager.c.widget["screensplit"]
    bar = manager.c.bar["top"]

    assert widget.info()["text"] == ""

    manager.c.next_layout()
    assert widget.info()["text"] == "top (max)"

    bar.fake_button_press(0, "top", 0, 0, 4)
    assert widget.info()["text"] == "bottom (columns)"

    bar.fake_button_press(0, "top", 0, 0, 4)
    assert widget.info()["text"] == "top (max)"

    bar.fake_button_press(0, "top", 0, 0, 5)
    assert widget.info()["text"] == "bottom (columns)"

    bar.fake_button_press(0, "top", 0, 0, 5)
    assert widget.info()["text"] == "top (max)"


@screensplit_config
def test_screensplit_text_format(manager):
    widget = manager.c.widget["screensplit_1"]
    manager.c.next_layout()
    assert widget.info()["text"] == "max - top"



# Copyright (c) 2008, Aldo Cortesi. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

import libqtile.config
from libqtile import layout
from libqtile.confreader import Config
from test.layouts.layout_utils import assert_focused


class FloatingConfig(Config):
    auto_fullscreen = True
    groups = [
        libqtile.config.Group("a"),
    ]
    layouts = [layout.Floating()]
    floating_layout = layout.Floating(
        fullscreen_border_width=15,
        max_border_width=10,
    )
    keys = []
    mouse = []
    screens = []
    follow_mouse_focus = False


floating_config = pytest.mark.parametrize("manager", [FloatingConfig], indirect=True)


@floating_config
def test_float_next_prev_window(manager):
    # spawn three windows
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")

    # focus previous windows
    assert_focused(manager, "three")
    manager.c.group.prev_window()
    assert_focused(manager, "two")
    manager.c.group.prev_window()
    assert_focused(manager, "one")
    # checking that it loops around properly
    manager.c.group.prev_window()
    assert_focused(manager, "three")

    # focus next windows
    # checking that it loops around properly
    manager.c.group.next_window()
    assert_focused(manager, "one")
    manager.c.group.next_window()
    assert_focused(manager, "two")
    manager.c.group.next_window()
    assert_focused(manager, "three")


@floating_config
def test_border_widths(manager):
    manager.test_window("one")

    # Default geometry
    info = manager.c.window.info()
    assert info["x"] == 350
    assert info["y"] == 250
    assert info["width"] == 100
    assert info["height"] == 100

    # Fullscreen
    manager.c.window.enable_fullscreen()
    info = manager.c.window.info()
    assert info["x"] == 0
    assert info["y"] == 0
    assert info["width"] == 770
    assert info["height"] == 570
    manager.c.window.disable_fullscreen()

    # Maximized
    manager.c.window.toggle_maximize()
    info = manager.c.window.info()
    assert info["x"] == 0
    assert info["y"] == 0
    assert info["width"] == 780
    assert info["height"] == 580



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

from libqtile.widget import idlerpg

# Set up some fake responses
# Default widget displayes "ttl" and "online" values but can also show
# any others keys from the response so we need to include an extra field
online_response = {"player": {"ttl": 1000, "online": "1", "unused": "0"}}
offline_response = {"player": {"ttl": 10300, "online": "0"}}


# The GenPollURL widget has been tested separately so we just need to test parse
# method of this widget
def test_idlerpg():
    idler = idlerpg.IdleRPG()
    assert idler.parse(online_response) == "IdleRPG: online TTL: 0:16:40"
    assert idler.parse(offline_response) == "IdleRPG: offline TTL: 2:51:40"



# Copyright (c) 2021 Jeroen Wijenbergh
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

import libqtile
from libqtile.command.base import expose_command
from libqtile.confreader import Config
from libqtile.layout.base import _SimpleLayoutBase


class DummyLayout(_SimpleLayoutBase):
    defaults = [
        ("current_offset", 0, ""),
        ("current_position", None, ""),
    ]

    def __init__(self, **config):
        _SimpleLayoutBase.__init__(self, **config)
        self.add_defaults(DummyLayout.defaults)

    def add_client(self, client):
        return super().add_client(
            client, offset_to_current=self.current_offset, client_position=self.current_position
        )

    def configure(self, client, screen_rect):
        pass

    @expose_command("up")
    def previous(self):
        _SimpleLayoutBase.previous()

    @expose_command("down")
    def next(self):
        _SimpleLayoutBase.next()


class BaseLayoutConfigBottom(Config):
    auto_fullscreen = True
    groups = [libqtile.config.Group("a")]
    layouts = [DummyLayout(current_position="bottom")]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []


class BaseLayoutConfigTop(Config):
    auto_fullscreen = True
    groups = [libqtile.config.Group("a")]
    layouts = [DummyLayout(current_position="top")]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []


baselayoutconfigbottom = pytest.mark.parametrize(
    "manager", [BaseLayoutConfigBottom], indirect=True
)
baselayoutconfigtop = pytest.mark.parametrize("manager", [BaseLayoutConfigTop], indirect=True)


@baselayoutconfigbottom
def test_base_client_position_bottom(manager):
    manager.test_window("one")
    manager.test_window("two")
    assert manager.c.layout.info()["clients"] == ["one", "two"]


@baselayoutconfigtop
def test_base_client_position_top(manager):
    manager.test_window("one")
    manager.test_window("two")
    assert manager.c.layout.info()["clients"] == ["two", "one"]



import wlroots.ffi_build as wlr
from cffi import FFI

from libqtile.backend.wayland.cffi import cairo_buffer, libinput

SOURCE = "\n".join(
    [
        wlr.SOURCE,
        libinput.SOURCE,
        cairo_buffer.SOURCE,
    ]
)

ffi = FFI()
ffi.set_source(
    "libqtile.backend.wayland._ffi",
    SOURCE,
    libraries=["wlroots", "input"],
    define_macros=[("WLR_USE_UNSTABLE", None)],
    include_dirs=[
        "/usr/include/pixman-1",
        "/usr/include/libdrm",
        wlr.include_dir,
    ],
)

ffi.include(wlr.ffi_builder)
ffi.cdef(libinput.CDEF)
ffi.cdef(cairo_buffer.CDEF)

if __name__ == "__main__":
    ffi.compile()



# Copyright (c) 2023, elParaguayo. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import libcst as cst
import libcst.matchers as m

from libqtile.scripts.migrations._base import (
    Check,
    MigrationTransformer,
    _QtileMigrator,
    add_migration,
)


class BluetoothArgsTransformer(MigrationTransformer):
    @m.call_if_inside(
        m.Call(func=m.Name("Bluetooth")) | m.Call(func=m.Attribute(attr=m.Name("Bluetooth")))
    )
    @m.leave(m.Arg(keyword=m.Name("hci")))
    def update_bluetooth_args(self, original_node, updated_node) -> cst.Arg:
        """Changes positional  argumentto 'widgets' kwargs."""
        self.lint(
            original_node,
            "The 'hci' argument is deprecated and should be replaced with 'device'.",
        )
        return updated_node.with_changes(keyword=cst.Name("device"))


class BluetoothArgs(_QtileMigrator):
    ID = "UpdateBluetoothArgs"

    SUMMARY = "Updates ``Bluetooth`` argument signature."

    HELP = """
    The ``Bluetooth`` widget previously accepted a ``hci`` keyword argument. This has
    been deprecated following a major overhaul of the widget and should be replaced with
    a keyword argument named ``device``.

    For example:

    .. code:: python

        widget.Bluetooth(hci="/dev_XX_XX_XX_XX_XX_XX")

    should be changed to:

    .. code::

        widget.Bluetooth(device="/dev_XX_XX_XX_XX_XX_XX")

    """

    AFTER_VERSION = "0.23.0"

    TESTS = [
        Check(
            """
            from libqtile import bar, widget
            from libqtile.widget import Bluetooth

            bar.Bar(
                [
                    Bluetooth(hci="/dev_XX_XX_XX_XX_XX_XX"),
                    widget.Bluetooth(hci="/dev_XX_XX_XX_XX_XX_XX"),
                ],
                20,
            )
            """,
            """
            from libqtile import bar, widget
            from libqtile.widget import Bluetooth

            bar.Bar(
                [
                    Bluetooth(device="/dev_XX_XX_XX_XX_XX_XX"),
                    widget.Bluetooth(device="/dev_XX_XX_XX_XX_XX_XX"),
                ],
                20,
            )
            """,
        ),
    ]

    visitor = BluetoothArgsTransformer


add_migration(BluetoothArgs)



# Copyright (c) 2017 Dirk Hartmann
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
from pathlib import Path

import pytest

import libqtile.config
import libqtile.layout
import libqtile.widget
from libqtile.confreader import Config
from test.helpers import Retry
from test.layouts.layout_utils import assert_focus_path, assert_focused


def spawn_cmd(title):
    script = Path(__file__).parent / "scripts" / "window.py"
    cmd = f"{sys.executable} {script.as_posix()} --name TestWindow {title} normal"
    return cmd


class ScratchPadBaseConfic(Config):
    auto_fullscreen = True
    screens = []
    groups = [
        libqtile.config.ScratchPad(
            "SCRATCHPAD",
            dropdowns=[
                libqtile.config.DropDown("dd-a", spawn_cmd("dd-a"), on_focus_lost_hide=False),
                libqtile.config.DropDown("dd-b", spawn_cmd("dd-b"), on_focus_lost_hide=False),
                libqtile.config.DropDown("dd-c", spawn_cmd("dd-c"), on_focus_lost_hide=True),
                libqtile.config.DropDown("dd-d", spawn_cmd("dd-d"), on_focus_lost_hide=True),
                libqtile.config.DropDown(
                    "dd-e",
                    spawn_cmd("dd-e"),
                    match=libqtile.config.Match(title="dd-e"),
                    on_focus_lost_hide=False,
                ),
            ],
        ),
        libqtile.config.ScratchPad(
            "SINGLE_SCRATCHPAD",
            dropdowns=[
                libqtile.config.DropDown("dd-e", spawn_cmd("dd-e"), on_focus_lost_hide=False),
                libqtile.config.DropDown("dd-f", spawn_cmd("dd-f"), on_focus_lost_hide=False),
            ],
            single=True,
        ),
        libqtile.config.Group("a"),
        libqtile.config.Group("b"),
    ]
    layouts = [libqtile.layout.max.Max()]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []


scratchpad_config = pytest.mark.parametrize("manager", [ScratchPadBaseConfic], indirect=True)


@Retry(ignore_exceptions=(KeyError,))
def is_spawned(manager, name, scratch_group="SCRATCHPAD"):
    manager.c.group[scratch_group].dropdown_info(name)["window"]
    return True


@Retry(ignore_exceptions=(ValueError,))
def is_killed(manager, name):
    if "window" not in manager.c.group["SCRATCHPAD"].dropdown_info(name):
        return True
    raise ValueError("not yet killed")


@scratchpad_config
def test_sratchpad_with_matcher(manager):
    # adjust command for current display
    manager.c.group["SCRATCHPAD"].dropdown_reconfigure("dd-e")

    manager.test_window("one")
    assert manager.c.group["a"].info()["windows"] == ["one"]

    # First toggling: wait for window
    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-e")
    is_spawned(manager, "dd-e")

    # assert window in current group "a"
    assert sorted(manager.c.group["a"].info()["windows"]) == ["dd-e", "one"]
    assert_focused(manager, "dd-e")

    # toggle again --> "hide" dd-e
    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-e")
    assert manager.c.group["a"].info()["windows"] == ["one"]
    assert_focused(manager, "one")
    assert manager.c.group["SCRATCHPAD"].info()["windows"] == ["dd-e"]

    # toggle again --> show again
    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-e")
    assert sorted(manager.c.group["a"].info()["windows"]) == ["dd-e", "one"]
    assert_focused(manager, "dd-e")
    assert manager.c.group["SCRATCHPAD"].info()["windows"] == []


@scratchpad_config
def test_toggling_single(manager):
    # adjust command for current display
    manager.c.group["SINGLE_SCRATCHPAD"].dropdown_reconfigure("dd-e")
    manager.c.group["SINGLE_SCRATCHPAD"].dropdown_reconfigure("dd-f")
    manager.c.group["SINGLE_SCRATCHPAD"].dropdown_reconfigure("dd-g")
    manager.c.group["SINGLE_SCRATCHPAD"].dropdown_reconfigure("dd-h")

    manager.test_window("one")
    assert manager.c.group["a"].info()["windows"] == ["one"]

    # First toggling: wait for window
    manager.c.group["SINGLE_SCRATCHPAD"].dropdown_toggle("dd-e")
    is_spawned(manager, "dd-e", "SINGLE_SCRATCHPAD")

    # assert window in current group "a"
    assert sorted(manager.c.group["a"].info()["windows"]) == ["dd-e", "one"]
    assert_focused(manager, "dd-e")

    # toggle another window, this should hide the previous one.
    manager.c.group["SINGLE_SCRATCHPAD"].dropdown_toggle("dd-f")
    is_spawned(manager, "dd-f", "SINGLE_SCRATCHPAD")
    assert sorted(manager.c.group["a"].info()["windows"]) == ["dd-f", "one"]
    assert_focused(manager, "dd-f")
    assert manager.c.group["SINGLE_SCRATCHPAD"].info()["windows"] == ["dd-e"]

    # toggle the scratchpad that is now visible.
    manager.c.group["SINGLE_SCRATCHPAD"].dropdown_toggle("dd-f")
    assert sorted(manager.c.group["a"].info()["windows"]) == ["one"]
    assert_focused(manager, "one")
    assert sorted(manager.c.group["SINGLE_SCRATCHPAD"].info()["windows"]) == ["dd-e", "dd-f"]


@scratchpad_config
def test_toggling(manager):
    manager.c.group["SCRATCHPAD"].dropdown_reconfigure("dd-a")

    manager.test_window("one")
    assert manager.c.group["a"].info()["windows"] == ["one"]

    # First toggling: wait for window
    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-a")
    is_spawned(manager, "dd-a")

    # assert window in current group "a"
    assert sorted(manager.c.group["a"].info()["windows"]) == ["dd-a", "one"]
    assert_focused(manager, "dd-a")

    # toggle again --> "hide" window in scratchpad group
    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-a")
    assert manager.c.group["a"].info()["windows"] == ["one"]
    assert_focused(manager, "one")
    assert manager.c.group["SCRATCHPAD"].info()["windows"] == ["dd-a"]

    # toggle again --> show again
    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-a")
    assert sorted(manager.c.group["a"].info()["windows"]) == ["dd-a", "one"]
    assert_focused(manager, "dd-a")
    assert manager.c.group["SCRATCHPAD"].info()["windows"] == []


@scratchpad_config
def test_focus_cycle(manager):
    manager.c.group["SCRATCHPAD"].dropdown_reconfigure("dd-a")
    manager.c.group["SCRATCHPAD"].dropdown_reconfigure("dd-b")

    manager.test_window("one")
    # spawn dd-a by toggling
    assert_focused(manager, "one")

    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-a")
    is_spawned(manager, "dd-a")
    assert_focused(manager, "dd-a")

    manager.test_window("two")
    assert_focused(manager, "two")

    # spawn dd-b by toggling
    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-b")
    is_spawned(manager, "dd-b")
    assert_focused(manager, "dd-b")

    # check all windows
    assert sorted(manager.c.group["a"].info()["windows"]) == ["dd-a", "dd-b", "one", "two"]

    assert_focus_path(manager, "one", "two", "dd-a", "dd-b")


@scratchpad_config
def test_focus_lost_hide(manager):
    manager.c.group["SCRATCHPAD"].dropdown_reconfigure("dd-c")
    manager.c.group["SCRATCHPAD"].dropdown_reconfigure("dd-d")

    manager.test_window("one")
    assert_focused(manager, "one")

    # spawn dd-c by toggling
    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-c")
    is_spawned(manager, "dd-c")
    assert_focused(manager, "dd-c")
    assert sorted(manager.c.group["a"].info()["windows"]) == ["dd-c", "one"]

    # New Window with Focus --> hide current DropDown
    manager.test_window("two")
    assert_focused(manager, "two")
    assert sorted(manager.c.group["a"].info()["windows"]) == ["one", "two"]
    assert sorted(manager.c.group["SCRATCHPAD"].info()["windows"]) == ["dd-c"]

    # spawn dd-b by toggling
    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-d")
    is_spawned(manager, "dd-d")
    assert_focused(manager, "dd-d")

    assert sorted(manager.c.group["a"].info()["windows"]) == ["dd-d", "one", "two"]
    assert sorted(manager.c.group["SCRATCHPAD"].info()["windows"]) == ["dd-c"]

    # focus next, is the first tiled window --> "hide" dd-d
    manager.c.group.next_window()
    assert_focused(manager, "one")
    assert sorted(manager.c.group["a"].info()["windows"]) == ["one", "two"]
    assert sorted(manager.c.group["SCRATCHPAD"].info()["windows"]) == ["dd-c", "dd-d"]

    # Bring dd-c to front
    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-c")
    assert_focused(manager, "dd-c")
    assert sorted(manager.c.group["a"].info()["windows"]) == ["dd-c", "one", "two"]
    assert sorted(manager.c.group["SCRATCHPAD"].info()["windows"]) == ["dd-d"]

    # Bring dd-d to front --> "hide dd-c
    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-d")
    assert_focused(manager, "dd-d")
    assert sorted(manager.c.group["a"].info()["windows"]) == ["dd-d", "one", "two"]
    assert sorted(manager.c.group["SCRATCHPAD"].info()["windows"]) == ["dd-c"]

    # change current group to "b" hids DropDowns
    manager.c.group["b"].toscreen()
    assert sorted(manager.c.group["a"].info()["windows"]) == ["one", "two"]
    assert sorted(manager.c.group["SCRATCHPAD"].info()["windows"]) == ["dd-c", "dd-d"]


@scratchpad_config
def test_kill(manager):
    manager.c.group["SCRATCHPAD"].dropdown_reconfigure("dd-a")

    manager.test_window("one")
    assert_focused(manager, "one")

    # dd-a has no window associated yet
    assert "window" not in manager.c.group["SCRATCHPAD"].dropdown_info("dd-a")

    # First toggling: wait for window
    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-a")
    is_spawned(manager, "dd-a")
    assert_focused(manager, "dd-a")
    assert manager.c.group["SCRATCHPAD"].dropdown_info("dd-a")["window"]["name"] == "dd-a"

    # kill current window "dd-a"
    manager.c.window.kill()
    manager.c.sync()
    is_killed(manager, "dd-a")
    assert_focused(manager, "one")
    assert "window" not in manager.c.group["SCRATCHPAD"].dropdown_info("dd-a")


@scratchpad_config
def test_floating_toggle(manager):
    manager.c.group["SCRATCHPAD"].dropdown_reconfigure("dd-a")

    manager.test_window("one")
    assert_focused(manager, "one")

    # dd-a has no window associated yet
    assert "window" not in manager.c.group["SCRATCHPAD"].dropdown_info("dd-a")
    # First toggling: wait for window
    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-a")
    is_spawned(manager, "dd-a")
    assert_focused(manager, "dd-a")

    assert "window" in manager.c.group["SCRATCHPAD"].dropdown_info("dd-a")
    assert sorted(manager.c.group["a"].info()["windows"]) == ["dd-a", "one"]

    manager.c.window.toggle_floating()
    # dd-a has no window associated any more, but is still in group
    assert "window" not in manager.c.group["SCRATCHPAD"].dropdown_info("dd-a")
    assert sorted(manager.c.group["a"].info()["windows"]) == ["dd-a", "one"]

    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-a")
    is_spawned(manager, "dd-a")
    assert sorted(manager.c.group["a"].info()["windows"]) == ["dd-a", "dd-a", "one"]


@scratchpad_config
def test_stepping_between_groups_should_skip_scratchpads(manager):
    # we are on a group
    manager.c.screen.next_group()
    # we are on b group
    manager.c.screen.next_group()
    # we should be on a group
    assert manager.c.group.info()["name"] == "a"

    manager.c.screen.prev_group()
    # we should be on b group
    assert manager.c.group.info()["name"] == "b"


@scratchpad_config
def test_skip_taskbar(manager):
    manager.c.group["SCRATCHPAD"].dropdown_reconfigure("dd-a")

    manager.test_window("one")
    assert_focused(manager, "one")

    # dd-a has no window associated yet
    assert "window" not in manager.c.group["SCRATCHPAD"].dropdown_info("dd-a")

    # First toggling: wait for window
    manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-a")
    is_spawned(manager, "dd-a")
    assert_focused(manager, "dd-a")
    assert manager.c.group["SCRATCHPAD"].dropdown_info("dd-a")["window"]["name"] == "dd-a"

    if manager.c.core.info()["backend"] == "x11":
        # check that window's _NET_WM_STATE contains _NET_WM_STATE_SKIP_TASKBAR
        net_wm_state = manager.c.window.eval("self.window.get_net_wm_state()")[1]
        assert "_NET_WM_STATE_SKIP_TASKBAR" in net_wm_state



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import pytest

from libqtile.core.lifecycle import Behavior, LifeCycle
from libqtile.log_utils import init_log


def fake_os_execv(executable, args):
    assert executable == "python"
    assert args == [
        "python",
        "arg1",
        "arg2",
        "--no-spawn",
        "--with-state=/tmp/test/fake_statefile",
    ]


def no_op(*args, **kwargs):
    pass


@pytest.fixture
def patched_lifecycle(monkeypatch):
    init_log()
    monkeypatch.setattr("libqtile.core.lifecycle.sys.executable", "python")
    monkeypatch.setattr("libqtile.core.lifecycle.sys.argv", ["arg1", "arg2"])
    monkeypatch.setattr("libqtile.core.lifecycle.atexit.register", no_op)
    monkeypatch.setattr("libqtile.core.lifecycle.os.execv", fake_os_execv)
    yield LifeCycle()


def test_restart_behaviour(patched_lifecycle, caplog):
    patched_lifecycle.behavior = Behavior.RESTART
    patched_lifecycle.state_file = "/tmp/test/fake_statefile"
    patched_lifecycle._atexit()
    assert caplog.record_tuples == [("libqtile", 30, "Restarting Qtile with os.execv(...)")]


def test_terminate_behavior(patched_lifecycle, caplog):
    patched_lifecycle.behavior = Behavior.TERMINATE
    patched_lifecycle._atexit()
    assert caplog.record_tuples == [("libqtile", 30, "Qtile will now terminate")]


def test_none_behavior(patched_lifecycle, caplog):
    patched_lifecycle.behavior = Behavior.NONE
    patched_lifecycle._atexit()
    assert caplog.record_tuples == []



# Copyright (c) 2008-2010 Aldo Cortesi
# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2011 Kenji_Takahashi
# Copyright (c) 2011 Paul Colomiets
# Copyright (c) 2012 roger
# Copyright (c) 2012 Craig Barnes
# Copyright (c) 2012-2015 Tycho Andersen
# Copyright (c) 2013 dequis
# Copyright (c) 2013 David R. Andersen
# Copyright (c) 2013 Tao Sauvage
# Copyright (c) 2014-2015 Sean Vig
# Copyright (c) 2014 Justin Bronder
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
import copy
import math
import subprocess
from typing import TYPE_CHECKING, Union

from libqtile import bar, configurable, confreader
from libqtile.command import interface
from libqtile.command.base import CommandError, CommandObject, expose_command
from libqtile.lazy import LazyCall
from libqtile.log_utils import logger
from libqtile.utils import ColorType, create_task

if TYPE_CHECKING:
    from typing import Any

    from libqtile.command.base import ItemT

# Each widget class must define which bar orientation(s) it supports by setting
# these bits in an 'orientations' class attribute. Simply having the attribute
# inherited by superclasses is discouraged, because if a superclass that was
# only supporting one orientation, adds support for the other, its subclasses
# will have to be adapted too, in general. ORIENTATION_NONE is only added for
# completeness' sake.
# +------------------------+--------------------+--------------------+
# | Widget bits            | Horizontal bar     | Vertical bar       |
# +========================+====================+====================+
# | ORIENTATION_NONE       | ConfigError raised | ConfigError raised |
# +------------------------+--------------------+--------------------+
# | ORIENTATION_HORIZONTAL | Widget displayed   | ConfigError raised |
# |                        | horizontally       |                    |
# +------------------------+--------------------+--------------------+
# | ORIENTATION_VERTICAL   | ConfigError raised | Widget displayed   |
# |                        |                    | vertically         |
# +------------------------+--------------------+--------------------+
# | ORIENTATION_BOTH       | Widget displayed   | Widget displayed   |
# |                        | horizontally       | vertically         |
# +------------------------+--------------------+--------------------+


class _Orientations(int):
    def __new__(cls, value, doc):
        return super().__new__(cls, value)

    def __init__(self, value, doc):
        self.doc = doc

    def __str__(self):
        return self.doc

    def __repr__(self):
        return self.doc


ORIENTATION_NONE = _Orientations(0, "none")
ORIENTATION_HORIZONTAL = _Orientations(1, "horizontal only")
ORIENTATION_VERTICAL = _Orientations(2, "vertical only")
ORIENTATION_BOTH = _Orientations(3, "horizontal and vertical")


class _Widget(CommandObject, configurable.Configurable):
    """Base Widget class

    If length is set to the special value `bar.STRETCH`, the bar itself will
    set the length to the maximum remaining space, after all other widgets have
    been configured.

    In horizontal bars, 'length' corresponds to the width of the widget; in
    vertical bars, it corresponds to the widget's height.

    The offsetx and offsety attributes are set by the Bar after all widgets
    have been configured.

    Callback functions can be assigned to button presses by passing a dict to the
    'callbacks' kwarg. No arguments are passed to the function so, if
    you need access to the qtile object, it needs to be imported into your code.

    ``lazy`` functions can also be passed as callback functions and can be used in
    the same way as keybindings.

    For example:

    .. code-block:: python

        from libqtile import qtile

        def open_calendar():
            qtile.spawn('gsimplecal next_month')

        clock = widget.Clock(
            mouse_callbacks={
                'Button1': open_calendar,
                'Button3': lazy.spawn('gsimplecal prev_month')
            }
        )

    When the clock widget receives a click with button 1, the ``open_calendar`` function
    will be executed.
    """

    orientations = ORIENTATION_BOTH

    # Default (empty set) is for all backends to be supported. Widgets can override this
    # to explicitly confirm which backends are supported
    supported_backends: set[str] = set()

    offsetx: int = 0
    offsety: int = 0
    defaults: list[tuple[str, Any, str]] = [
        ("background", None, "Widget background color"),
        (
            "mouse_callbacks",
            {},
            "Dict of mouse button press callback functions. Accepts functions and ``lazy`` calls.",
        ),
    ]

    def __init__(self, length, **config):
        """
        length: bar.STRETCH, bar.CALCULATED, or a specified length.
        """
        CommandObject.__init__(self)
        self.name = self.__class__.__name__.lower()
        if "name" in config:
            self.name = config["name"]

        configurable.Configurable.__init__(self, **config)
        self.add_defaults(_Widget.defaults)

        if length in (bar.CALCULATED, bar.STRETCH):
            self.length_type = length
            self.length = 0
        elif isinstance(length, int):
            self.length_type = bar.STATIC
            self.length = length
        else:
            raise confreader.ConfigError("Widget width must be an int")

        self.configured = False
        self._futures: list[asyncio.Handle] = []
        self._mirrors: set[_Widget] = set()
        self.finalized = False

    @property
    def length(self):
        if self.length_type == bar.CALCULATED:
            return int(self.calculate_length())
        return self._length

    @length.setter
    def length(self, value):
        self._length = value

    @property
    def width(self):
        if self.bar.horizontal:
            return self.length
        return self.bar.width

    @property
    def height(self):
        if self.bar.horizontal:
            return self.bar.height
        return self.length

    @property
    def offset(self):
        if self.bar.horizontal:
            return self.offsetx
        return self.offsety

    def _test_orientation_compatibility(self, horizontal):
        if horizontal:
            if not self.orientations & ORIENTATION_HORIZONTAL:
                raise confreader.ConfigError(
                    self.__class__.__name__
                    + " is not compatible with the orientation of the bar."
                )
        elif not self.orientations & ORIENTATION_VERTICAL:
            raise confreader.ConfigError(
                self.__class__.__name__ + " is not compatible with the orientation of the bar."
            )

    def timer_setup(self):
        """This is called exactly once, after the widget has been configured
        and timers are available to be set up."""
        pass

    def _configure(self, qtile, bar):
        self._test_orientation_compatibility(bar.horizontal)

        self.qtile = qtile
        self.bar = bar
        self.drawer = bar.window.create_drawer(self.bar.width, self.bar.height)

        # Clear this flag as widget may be restarted (e.g. if screen removed and re-added)
        self.finalized = False

        # Timers are added to futures list so they can be cancelled if the `finalize` method is
        # called before the timers have fired.
        if not self.configured:
            timer = self.qtile.call_soon(self.timer_setup)
            async_timer = self.qtile.call_soon(asyncio.create_task, self._config_async())

            # Add these to our list of futures so they can be cancelled.
            self._futures.extend([timer, async_timer])

    async def _config_async(self):
        """
        This is called once when the main eventloop has started. this
        happens after _configure has been run.

        Widgets that need to use asyncio coroutines after this point may
        wish to initialise the relevant code (e.g. connections to dbus
        using dbus_next) here.
        """
        pass

    def finalize(self):
        for future in self._futures:
            future.cancel()
        if hasattr(self, "layout") and self.layout:
            self.layout.finalize()
        self.drawer.finalize()
        self.finalized = True

        # Reset configuration status so the widget can be reconfigured
        # e.g. when screen is re-added
        self.configured = False

    def clear(self):
        self.drawer.set_source_rgb(self.bar.background)
        self.drawer.fillrect(self.offsetx, self.offsety, self.width, self.height)

    @expose_command()
    def info(self):
        """Info for this object."""
        return dict(
            name=self.name,
            offset=self.offset,
            length=self.length,
            width=self.width,
            height=self.height,
        )

    def add_callbacks(self, defaults):
        """Add default callbacks with a lower priority than user-specified callbacks."""
        defaults.update(self.mouse_callbacks)
        self.mouse_callbacks = defaults

    def button_press(self, x, y, button):
        name = "Button{0}".format(button)
        if name in self.mouse_callbacks:
            cmd = self.mouse_callbacks[name]
            if isinstance(cmd, LazyCall):
                if cmd.check(self.qtile):
                    status, val = self.qtile.server.call(
                        (cmd.selectors, cmd.name, cmd.args, cmd.kwargs, False)
                    )
                    if status in (interface.ERROR, interface.EXCEPTION):
                        logger.error("Mouse callback command error %s: %s", cmd.name, val)
            else:
                cmd()

    def button_release(self, x, y, button):
        pass

    def get(self, q, name):
        """
        Utility function for quick retrieval of a widget by name.
        """
        w = q.widgets_map.get(name)
        if not w:
            raise CommandError("No such widget: %s" % name)
        return w

    def _items(self, name: str) -> ItemT:
        if name == "bar":
            return True, []
        elif name == "screen":
            return True, []
        return None

    def _select(self, name, sel):
        if name == "bar":
            return self.bar
        elif name == "screen":
            return self.bar.screen

    def draw(self):
        """
        Method that draws the widget. You may call this explicitly to
        redraw the widget, but only if the length of the widget hasn't
        changed. If it has, you must call bar.draw instead.
        """
        raise NotImplementedError

    def calculate_length(self):
        """
        Must be implemented if the widget can take CALCULATED for length.
        It must return the width of the widget if it's installed in a
        horizontal bar; it must return the height of the widget if it's
        installed in a vertical bar. Usually you will test the orientation
        of the bar with 'self.bar.horizontal'.
        """
        raise NotImplementedError

    def timeout_add(self, seconds, method, method_args=()):
        """
        This method calls ``.call_later`` with given arguments.
        """
        # Don't add timers for finalised widgets
        if self.finalized:
            return

        future = self.qtile.call_later(seconds, self._wrapper, method, *method_args)

        self._futures.append(future)
        return future

    def call_process(self, command, **kwargs):
        """
        This method uses `subprocess.check_output` to run the given command
        and return the string from stdout, which is decoded when using
        Python 3.
        """
        return subprocess.check_output(command, **kwargs, encoding="utf-8")

    def _remove_dead_timers(self):
        """Remove completed and cancelled timers from the list."""

        def is_ready(timer):
            return timer in self.qtile._eventloop._ready

        self._futures = [
            timer
            for timer in self._futures
            # Filter out certain handles...
            if not (
                timer.cancelled()
                # Once a scheduled timer is ready to be run its _scheduled flag is set to False
                # and it's added to the loop's `_ready` queue
                or (
                    isinstance(timer, asyncio.TimerHandle)
                    and not timer._scheduled
                    and not is_ready(timer)
                )
                # Callbacks scheduled via `call_soon` are put into the loop's `_ready` queue
                # and are removed once they've been executed
                or (isinstance(timer, asyncio.Handle) and not is_ready(timer))
            )
        ]

    def _wrapper(self, method, *method_args):
        self._remove_dead_timers()
        try:
            if asyncio.iscoroutinefunction(method):
                create_task(method(*method_args))
            elif asyncio.iscoroutine(method):
                create_task(method)
            else:
                method(*method_args)
        except:  # noqa: E722
            logger.exception("got exception from widget timer")

    def create_mirror(self):
        return Mirror(self, background=self.background)

    def clone(self):
        return copy.deepcopy(self)

    def mouse_enter(self, x, y):
        pass

    def mouse_leave(self, x, y):
        pass

    def _draw_with_mirrors(self) -> None:
        self._old_draw()
        for mirror in self._mirrors:
            if not mirror.configured:
                continue

            # If the widget and mirror are on the same bar then we could have an
            # infinite loop when we call bar.draw(). mirror.draw() will trigger a resize
            # if it's the wrong size.
            if mirror.length_type == bar.CALCULATED and mirror.bar is not self.bar:
                mirror.bar.draw()
            else:
                mirror.draw()

    def add_mirror(self, widget: _Widget):
        if not self._mirrors:
            self._old_draw = self.draw
            self.draw = self._draw_with_mirrors  # type: ignore

        self._mirrors.add(widget)
        if not self.drawer.has_mirrors:
            self.drawer.has_mirrors = True

    def remove_mirror(self, widget: _Widget):
        try:
            self._mirrors.remove(widget)
        except KeyError:
            pass

        if not self._mirrors:
            self.drawer.has_mirrors = False

            if hasattr(self, "_old_draw"):
                # Deletes the reference to draw and falls back to the original
                del self.draw
                del self._old_draw


class _TextBox(_Widget):
    """
    Base class for widgets that are just boxes containing text.
    """

    orientations = ORIENTATION_BOTH
    defaults = [
        ("font", "sans", "Default font"),
        ("fontsize", None, "Font size. Calculated if None."),
        ("padding", None, "Padding. Calculated if None."),
        ("foreground", "ffffff", "Foreground colour"),
        ("fontshadow", None, "font shadow color, default is None(no shadow)"),
        ("markup", True, "Whether or not to use pango markup"),
        (
            "fmt",
            "{}",
            "Format to apply to the string returned by the widget. Main purpose: applying markup. "
            "For a widget that returns ``foo``, using ``fmt='<i>{}</i>'`` would give you ``<i>foo</i>``. "
            "To control what the widget outputs in the first place, use the ``format`` paramater of the widget (if it has one).",
        ),
        ("max_chars", 0, "Maximum number of characters to display in widget."),
        (
            "scroll",
            False,
            "Whether text should be scrolled. When True, you must set the widget's ``width``.",
        ),
        (
            "scroll_repeat",
            True,
            "Whether text should restart scrolling once the text has ended",
        ),
        (
            "scroll_delay",
            2,
            "Number of seconds to pause before starting scrolling and restarting/clearing text at end",
        ),
        ("scroll_step", 1, "Number of pixels to scroll with each step"),
        ("scroll_interval", 0.1, "Time in seconds before next scrolling step"),
        (
            "scroll_clear",
            False,
            "Whether text should scroll completely away (True) or stop when the end of the text is shown (False)",
        ),
        ("scroll_hide", False, "Whether the widget should hide when scrolling has finished"),
        (
            "scroll_fixed_width",
            False,
            "When ``scroll=True`` the ``width`` parameter is a maximum width and, when text is shorter than this, the widget will resize. "
            "Setting ``scroll_fixed_width=True`` will force the widget to have a fixed width, regardless of the size of the text.",
        ),
    ]  # type: list[tuple[str, Any, str]]

    def __init__(self, text=" ", width=bar.CALCULATED, **config):
        self.layout = None
        _Widget.__init__(self, width, **config)
        self.add_defaults(_TextBox.defaults)
        self.text = text
        self._is_scrolling = False
        self._should_scroll = False
        self._scroll_offset = 0
        self._scroll_queued = False
        self._scroll_timer = None
        self._scroll_width = width

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        if len(value) > self.max_chars > 0:
            value = value[: self.max_chars] + "…"
        self._text = value
        if self.layout:
            self.layout.text = self.formatted_text
            if self.scroll:
                self.check_width()
                self.reset_scroll()

    @property
    def formatted_text(self):
        return self.fmt.format(self._text)

    @property
    def foreground(self):
        return self._foreground

    @foreground.setter
    def foreground(self, fg):
        self._foreground = fg
        if self.layout:
            self.layout.colour = fg

    @property
    def font(self):
        return self._font

    @font.setter
    def font(self, value):
        self._font = value
        if self.layout:
            self.layout.font = value

    @property
    def fontshadow(self):
        return self._fontshadow

    @fontshadow.setter
    def fontshadow(self, value):
        self._fontshadow = value
        if self.layout:
            self.layout.font_shadow = value

    @property
    def actual_padding(self):
        if self.padding is None:
            return self.fontsize / 2
        else:
            return self.padding

    def _configure(self, qtile, bar):
        _Widget._configure(self, qtile, bar)
        if self.fontsize is None:
            self.fontsize = self.bar.height - self.bar.height / 5
        self.layout = self.drawer.textlayout(
            self.formatted_text,
            self.foreground,
            self.font,
            self.fontsize,
            self.fontshadow,
            markup=self.markup,
        )
        if not isinstance(self._scroll_width, int) and self.scroll:
            logger.warning("%s: You must specify a width when enabling scrolling.", self.name)
            self.scroll = False

        if self.scroll:
            self.check_width()

    def check_width(self):
        """
        Check whether the widget needs to have calculated or fixed width
        and whether the text should be scrolled.
        """
        if self.layout.width > self._scroll_width:
            self.length_type = bar.STATIC
            self.length = self._scroll_width
            self._is_scrolling = True
            self._should_scroll = True
        else:
            if self.scroll_fixed_width:
                self.length_type = bar.STATIC
                self.length = self._scroll_width
            else:
                self.length_type = bar.CALCULATED
            self._should_scroll = False

    def calculate_length(self):
        if self.text:
            if self.bar.horizontal:
                return min(self.layout.width, self.bar.width) + self.actual_padding * 2
            else:
                return min(self.layout.width, self.bar.height) + self.actual_padding * 2
        else:
            return 0

    def can_draw(self):
        can_draw = (
            self.layout is not None and not self.layout.finalized() and self.offsetx is not None
        )  # if the bar hasn't placed us yet
        return can_draw

    def draw(self):
        if not self.can_draw():
            return
        self.drawer.clear(self.background or self.bar.background)

        # size = self.bar.height if self.bar.horizontal else self.bar.width
        self.drawer.ctx.save()

        if not self.bar.horizontal:
            # Left bar reads bottom to top
            if self.bar.screen.left is self.bar:
                self.drawer.ctx.rotate(-90 * math.pi / 180.0)
                self.drawer.ctx.translate(-self.length, 0)

            # Right bar is top to bottom
            else:
                self.drawer.ctx.translate(self.bar.width, 0)
                self.drawer.ctx.rotate(90 * math.pi / 180.0)

        # If we're scrolling, we clip the context to the scroll width less the padding
        # Move the text layout position (and we only see the clipped portion)
        if self._should_scroll:
            self.drawer.ctx.rectangle(
                self.actual_padding,
                0,
                self._scroll_width - 2 * self.actual_padding,
                self.bar.size,
            )
            self.drawer.ctx.clip()

        size = self.bar.height if self.bar.horizontal else self.bar.width

        self.layout.draw(
            (self.actual_padding or 0) - self._scroll_offset,
            int(size / 2.0 - self.layout.height / 2.0) + 1,
        )
        self.drawer.ctx.restore()

        self.drawer.draw(
            offsetx=self.offsetx, offsety=self.offsety, width=self.width, height=self.height
        )

        # We only want to scroll if:
        # - User has asked us to scroll and the scroll width is smaller than the layout (should_scroll=True)
        # - We are still scrolling (is_scrolling=True)
        # - We haven't already queued the next scroll (scroll_queued=False)
        if self._should_scroll and self._is_scrolling and not self._scroll_queued:
            self._scroll_queued = True
            if self._scroll_offset == 0:
                interval = self.scroll_delay
            else:
                interval = self.scroll_interval
            self._scroll_timer = self.timeout_add(interval, self.do_scroll)

    def do_scroll(self):
        # Allow the next scroll tick to be queued
        self._scroll_queued = False

        # If we're still scrolling, adjust the next offset
        if self._is_scrolling:
            self._scroll_offset += self.scroll_step

        # Check whether we need to stop scrolling when:
        # - we've scrolled all the text off the widget (scroll_clear = True)
        # - the final pixel is visible (scroll_clear = False)
        if (self.scroll_clear and self._scroll_offset > self.layout.width) or (
            not self.scroll_clear
            and (self.layout.width - self._scroll_offset)
            < (self._scroll_width - 2 * self.actual_padding)
        ):
            self._is_scrolling = False

        # We've reached the end of the scroll so what next?
        if not self._is_scrolling:
            if self.scroll_repeat:
                # Pause and restart scrolling
                self._scroll_timer = self.timeout_add(self.scroll_delay, self.reset_scroll)
            elif self.scroll_hide:
                # Clear the text
                self._scroll_timer = self.timeout_add(self.scroll_delay, self.hide_scroll)
            # If neither of these options then the text is no longer updated.

        self.draw()

    def reset_scroll(self):
        self._scroll_offset = 0
        self._is_scrolling = True
        self._scroll_queued = False
        if self._scroll_timer:
            self._scroll_timer.cancel()
        self.draw()

    def hide_scroll(self):
        self.update("")

    @expose_command()
    def set_font(
        self,
        font: Union[str, None] = None,
        fontsize: int = 0,
        fontshadow: ColorType = "",
    ):
        """
        Change the font used by this widget. If font is None, the current
        font is used.
        """
        if font is not None:
            self.font = font
        if fontsize != 0:
            self.fontsize = fontsize
        if fontshadow != "":
            self.fontshadow = fontshadow
        self.bar.draw()

    @expose_command()
    def info(self):
        d = _Widget.info(self)
        d["foreground"] = self.foreground
        d["text"] = self.formatted_text
        return d

    def update(self, text):
        """Update the widget text."""
        # Don't try to update text in dead layouts
        # This is mainly required for ThreadPoolText based widgets as the
        # polling function cannot be cancelled and so may be called after the widget
        # is finalised.
        if not self.can_draw():
            return

        if self.text == text:
            return
        if text is None:
            text = ""

        old_width = self.layout.width
        self.text = text

        # If our width hasn't changed, we just draw ourselves. Otherwise,
        # we draw the whole bar.
        if self.layout.width == old_width:
            self.draw()
        else:
            self.bar.draw()


class InLoopPollText(_TextBox):
    """A common interface for polling some 'fast' information, munging it, and
    rendering the result in a text box. You probably want to use
    ThreadPoolText instead.

    ('fast' here means that this runs /in/ the event loop, so don't block! If
    you want to run something nontrivial, use ThreadedPollWidget.)"""

    defaults = [
        (
            "update_interval",
            600,
            "Update interval in seconds, if none, the widget updates only once.",
        ),
    ]  # type: list[tuple[str, Any, str]]

    def __init__(self, default_text="N/A", **config):
        _TextBox.__init__(self, default_text, **config)
        self.add_defaults(InLoopPollText.defaults)

    def timer_setup(self):
        update_interval = self.tick()
        # If self.update_interval is defined and .tick() returns None, re-call
        # after self.update_interval
        if update_interval is None and self.update_interval is not None:
            self.timeout_add(self.update_interval, self.timer_setup)
        # We can change the update interval by returning something from .tick()
        elif update_interval:
            self.timeout_add(update_interval, self.timer_setup)
        # If update_interval is False, we won't re-call

    def _configure(self, qtile, bar):
        should_tick = self.configured
        _TextBox._configure(self, qtile, bar)

        # Update when we are being re-configured.
        if should_tick:
            self.tick()

    def button_press(self, x, y, button):
        self.tick()
        _TextBox.button_press(self, x, y, button)

    def poll(self):
        return "N/A"

    def tick(self):
        text = self.poll()
        self.update(text)


class ThreadPoolText(_TextBox):
    """A common interface for wrapping blocking events which when triggered
    will update a textbox.

    The poll method is intended to wrap a blocking function which may take
    quite a while to return anything.  It will be executed as a future and
    should return updated text when completed.  It may also return None to
    disable any further updates.

    param: text - Initial text to display.
    """

    defaults = [
        (
            "update_interval",
            600,
            "Update interval in seconds, if none, the widget updates only once.",
        ),
    ]  # type: list[tuple[str, Any, str]]

    def __init__(self, text, **config):
        super().__init__(text, **config)
        self.add_defaults(ThreadPoolText.defaults)

    def timer_setup(self):
        def on_done(future):
            try:
                result = future.result()
            except Exception:
                result = None
                logger.exception("poll() raised exceptions, not rescheduling")

            if result is not None:
                try:
                    self.update(result)

                    if self.update_interval is not None:
                        self.timeout_add(self.update_interval, self.timer_setup)

                except Exception:
                    logger.exception("Failed to reschedule timer for %s.", self.name)
            else:
                logger.warning("%s's poll() returned None, not rescheduling", self.name)

        self.future = self.qtile.run_in_executor(self.poll)
        self.future.add_done_callback(on_done)

    def poll(self):
        pass

    @expose_command()
    def force_update(self):
        """Immediately poll the widget. Existing timers are unaffected."""
        self.update(self.poll())


# these two classes below look SUSPICIOUSLY similar


class PaddingMixin(configurable.Configurable):
    """Mixin that provides padding(_x|_y|)

    To use it, subclass and add this to __init__:

        self.add_defaults(base.PaddingMixin.defaults)
    """

    defaults = [
        ("padding", 3, "Padding inside the box"),
        ("padding_x", None, "X Padding. Overrides 'padding' if set"),
        ("padding_y", None, "Y Padding. Overrides 'padding' if set"),
    ]  # type: list[tuple[str, Any, str]]

    padding_x = configurable.ExtraFallback("padding_x", "padding")
    padding_y = configurable.ExtraFallback("padding_y", "padding")


class MarginMixin(configurable.Configurable):
    """Mixin that provides margin(_x|_y|)

    To use it, subclass and add this to __init__:

        self.add_defaults(base.MarginMixin.defaults)
    """

    defaults = [
        ("margin", 3, "Margin inside the box"),
        ("margin_x", None, "X Margin. Overrides 'margin' if set"),
        ("margin_y", None, "Y Margin. Overrides 'margin' if set"),
    ]  # type: list[tuple[str, Any, str]]

    margin_x = configurable.ExtraFallback("margin_x", "margin")
    margin_y = configurable.ExtraFallback("margin_y", "margin")


class Mirror(_Widget):
    """
    A widget for showing the same widget content in more than one place, for
    instance, on bars across multiple screens.

    You don't need to use it directly; instead, just instantiate your widget
    once and hand it in to multiple bars. For instance::

        cpu = widget.CPUGraph()
        clock = widget.Clock()

        screens = [
            Screen(top=bar.Bar([widget.GroupBox(), cpu, clock])),
            Screen(top=bar.Bar([widget.GroupBox(), cpu, clock])),
        ]

    Widgets can be passed to more than one bar, so that there don't need to be
    any duplicates executing the same code all the time, and they'll always be
    visually identical.

    This works for all widgets that use `drawers` (and nothing else) to display
    their contents. Currently, this is all widgets except for `Systray`.
    """

    def __init__(self, reflection, **config):
        _Widget.__init__(self, reflection.length, **config)
        self.reflects = reflection
        self._length = 0
        self.length_type = self.reflects.length_type
        if self.length_type is bar.STATIC:
            self._length = self.reflects._length

    def _configure(self, qtile, bar):
        _Widget._configure(self, qtile, bar)
        self.reflects.add_mirror(self)
        # We need to fill the background once before `draw` is called so, if
        # there's no reflection, the mirror matches its parent bar.
        self.drawer.clear(self.background or self.bar.background)

    def calculate_length(self):
        return self.reflects.calculate_length()

    @property
    def length(self):
        if self.length_type != bar.STRETCH:
            return self.reflects.length
        return self._length

    @length.setter
    def length(self, value):
        self._length = value

    def draw(self):
        if self.length <= 0:
            return
        self.drawer.clear_rect()
        self.reflects.drawer.paint_to(self.drawer)
        self.drawer.draw(offsetx=self.offset, offsety=self.offsety, width=self.width)

    def button_press(self, x, y, button):
        self.reflects.button_press(x, y, button)

    def mouse_enter(self, x, y):
        self.reflects.mouse_enter(x, y)

    def mouse_leave(self, x, y):
        self.reflects.mouse_leave(x, y)

    def finalize(self):
        self.reflects.remove_mirror(self)
        _Widget.finalize(self)



# Copyright (c) 2013 Jacob Mourelos
# Copyright (c) 2014 Shepilov Vladislav
# Copyright (c) 2014-2015 Sean Vig
# Copyright (c) 2014 Tycho Andersen
# Copyright (c) 2019 zordsdavini
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import re
from abc import ABCMeta, abstractmethod
from pathlib import Path
from subprocess import CalledProcessError, check_output
from typing import TYPE_CHECKING

from libqtile.command.base import expose_command
from libqtile.confreader import ConfigError
from libqtile.log_utils import logger
from libqtile.widget import base

if TYPE_CHECKING:
    from libqtile.core.manager import Qtile


class _BaseLayoutBackend(metaclass=ABCMeta):
    def __init__(self, qtile: Qtile):
        """
        This handles getting and setter the keyboard layout with the appropriate
        backend.
        """

    @abstractmethod
    def get_keyboard(self) -> str:
        """
        Return the currently used keyboard layout as a string

        Examples: "us", "us dvorak".  In case of error returns "unknown".
        """

    def set_keyboard(self, layout: str, options: str | None) -> None:
        """
        Set the keyboard layout with specified options.
        """


class _X11LayoutBackend(_BaseLayoutBackend):
    kb_layout_regex = re.compile(r"layout:\s+(?P<layout>[\w-]+)")
    kb_variant_regex = re.compile(r"variant:\s+(?P<variant>[\w-]+)")

    def get_keyboard(self) -> str:
        try:
            command = "setxkbmap -verbose 10 -query"
            setxkbmap_output = check_output(command.split(" ")).decode()
        except CalledProcessError:
            logger.exception("Can not get the keyboard layout:")
            return "unknown"
        except OSError:
            logger.exception("Please, check that xset is available:")
            return "unknown"

        match_layout = self.kb_layout_regex.search(setxkbmap_output)
        if match_layout is None:
            return "ERR"
        keyboard = match_layout.group("layout")

        match_variant = self.kb_variant_regex.search(setxkbmap_output)
        if match_variant:
            keyboard += " " + match_variant.group("variant")
        return keyboard

    def set_keyboard(self, layout: str, options: str | None) -> None:
        command = ["setxkbmap"]
        command.extend(layout.split(" "))
        if options:
            command.extend(["-option", options])
        try:
            check_output(command)
        except CalledProcessError:
            logger.error("Cannot change the keyboard layout.")
        except OSError:
            logger.error("Please, check that setxkbmap is available.")
        else:
            # Load Xmodmap if it's available
            if Path("~/.Xmodmap").expanduser().is_file():
                try:
                    check_output("xmodmap $HOME/.Xmodmap", shell=True)
                except CalledProcessError:
                    logger.error("Could not load ~/.Xmodmap.")


class _WaylandLayoutBackend(_BaseLayoutBackend):
    def __init__(self, qtile: Qtile) -> None:
        self.set_keymap = qtile.core.set_keymap
        self._layout: str = ""

    def get_keyboard(self) -> str:
        return self._layout

    def set_keyboard(self, layout: str, options: str | None) -> None:
        maybe_variant: str | None = None
        if " " in layout:
            layout_name, maybe_variant = layout.split(" ", maxsplit=1)
        else:
            layout_name = layout
        self.set_keymap(layout_name, options, maybe_variant)
        self._layout = layout


layout_backends = {
    "x11": _X11LayoutBackend,
    "wayland": _WaylandLayoutBackend,
}


class KeyboardLayout(base.InLoopPollText):
    """Widget for changing and displaying the current keyboard layout

    To use this widget effectively you need to specify keyboard layouts you want to use
    (using "configured_keyboards") and bind function "next_keyboard" to specific keys in
    order to change layouts.

    For example:

        Key([mod], "space", lazy.widget["keyboardlayout"].next_keyboard(), desc="Next keyboard layout."),

    When running Qtile with the X11 backend, this widget requires setxkbmap to be available.
    Xmodmap will also be used if .Xmodmap file is available.
    """

    defaults = [
        ("update_interval", 1, "Update time in seconds."),
        (
            "configured_keyboards",
            ["us"],
            "A list of predefined keyboard layouts "
            "represented as strings. For example: "
            "['us', 'us colemak', 'es', 'fr'].",
        ),
        (
            "display_map",
            {},
            "Custom display of layout. Key should be in format "
            "'layout variant'. For example: "
            "{'us': 'us', 'lt sgs': 'sgs', 'ru phonetic': 'ru'}",
        ),
        ("option", None, "string of setxkbmap option. Ex., 'compose:menu,grp_led:scroll'"),
    ]

    def __init__(self, **config):
        base.InLoopPollText.__init__(self, **config)
        self.add_defaults(KeyboardLayout.defaults)
        self.add_callbacks({"Button1": self.next_keyboard})

    def _configure(self, qtile, bar):
        base.InLoopPollText._configure(self, qtile, bar)

        if qtile.core.name not in layout_backends:
            raise ConfigError("KeyboardLayout does not support backend: " + qtile.core.name)

        self.backend = layout_backends[qtile.core.name](qtile)
        self.backend.set_keyboard(self.configured_keyboards[0], self.option)

    @expose_command()
    def next_keyboard(self):
        """set the next layout in the list of configured keyboard layouts as
        new current layout in use

        If the current keyboard layout is not in the list, it will set as new
        layout the first one in the list.
        """

        current_keyboard = self.backend.get_keyboard()
        if current_keyboard in self.configured_keyboards:
            # iterate the list circularly
            next_keyboard = self.configured_keyboards[
                (self.configured_keyboards.index(current_keyboard) + 1)
                % len(self.configured_keyboards)
            ]
        else:
            next_keyboard = self.configured_keyboards[0]

        self.backend.set_keyboard(next_keyboard, self.option)

        self.tick()

    def poll(self):
        keyboard = self.backend.get_keyboard()
        if keyboard in self.display_map.keys():
            return self.display_map[keyboard]
        return keyboard.upper()



# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2012, 2014-2015 Tycho Andersen
# Copyright (c) 2013 Mattias Svala
# Copyright (c) 2013 Craig Barnes
# Copyright (c) 2014 ramnes
# Copyright (c) 2014 Sean Vig
# Copyright (c) 2014 Adi Sieker
# Copyright (c) 2014 Chris Wesseling
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from time import sleep

import pytest

import libqtile.config
from libqtile import layout
from libqtile.confreader import Config
from test.layouts.layout_utils import assert_focus_path, assert_focused


class RatioTileConfig(Config):
    auto_fullscreen = True
    groups = [
        libqtile.config.Group("a"),
        libqtile.config.Group("b"),
        libqtile.config.Group("c"),
        libqtile.config.Group("d"),
    ]
    layouts = [layout.RatioTile(ratio=0.5), layout.RatioTile(), layout.RatioTile(fancy=True)]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []
    follow_mouse_focus = False


ratiotile_config = pytest.mark.parametrize("manager", [RatioTileConfig], indirect=True)


@ratiotile_config
def test_ratiotile_add_windows(manager):
    for i in range(12):
        manager.test_window(str(i))
        if i == 0:
            assert manager.c.layout.info()["layout_info"] == [(0, 0, 800, 600)]
        elif i == 1:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 400, 600),
                (400, 0, 400, 600),
            ]
        elif i == 2:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 266, 600),
                (266, 0, 266, 600),
                (532, 0, 268, 600),
            ]
        elif i == 3:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 200, 600),
                (200, 0, 200, 600),
                (400, 0, 200, 600),
                (600, 0, 200, 600),
            ]
        elif i == 4:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 160, 600),
                (160, 0, 160, 600),
                (320, 0, 160, 600),
                (480, 0, 160, 600),
                (640, 0, 160, 600),
            ]
        elif i == 5:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 133, 600),
                (133, 0, 133, 600),
                (266, 0, 133, 600),
                (399, 0, 133, 600),
                (532, 0, 133, 600),
                (665, 0, 135, 600),
            ]
        elif i == 6:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 200, 300),
                (200, 0, 200, 300),
                (400, 0, 200, 300),
                (600, 0, 200, 300),
                (0, 300, 266, 300),
                (266, 300, 266, 300),
                (532, 300, 268, 300),
            ]
        elif i == 7:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 200, 300),
                (200, 0, 200, 300),
                (400, 0, 200, 300),
                (600, 0, 200, 300),
                (0, 300, 200, 300),
                (200, 300, 200, 300),
                (400, 300, 200, 300),
                (600, 300, 200, 300),
            ]
        elif i == 8:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 160, 300),
                (160, 0, 160, 300),
                (320, 0, 160, 300),
                (480, 0, 160, 300),
                (640, 0, 160, 300),
                (0, 300, 200, 300),
                (200, 300, 200, 300),
                (400, 300, 200, 300),
                (600, 300, 200, 300),
            ]
        elif i == 9:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 160, 300),
                (160, 0, 160, 300),
                (320, 0, 160, 300),
                (480, 0, 160, 300),
                (640, 0, 160, 300),
                (0, 300, 160, 300),
                (160, 300, 160, 300),
                (320, 300, 160, 300),
                (480, 300, 160, 300),
                (640, 300, 160, 300),
            ]
        elif i == 10:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 133, 300),
                (133, 0, 133, 300),
                (266, 0, 133, 300),
                (399, 0, 133, 300),
                (532, 0, 133, 300),
                (665, 0, 135, 300),
                (0, 300, 160, 300),
                (160, 300, 160, 300),
                (320, 300, 160, 300),
                (480, 300, 160, 300),
                (640, 300, 160, 300),
            ]
        elif i == 11:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 133, 300),
                (133, 0, 133, 300),
                (266, 0, 133, 300),
                (399, 0, 133, 300),
                (532, 0, 133, 300),
                (665, 0, 135, 300),
                (0, 300, 133, 300),
                (133, 300, 133, 300),
                (266, 300, 133, 300),
                (399, 300, 133, 300),
                (532, 300, 133, 300),
                (665, 300, 135, 300),
            ]
        else:
            assert False


@ratiotile_config
def test_ratiotile_add_windows_golden_ratio(manager):
    manager.c.next_layout()
    for i in range(12):
        manager.test_window(str(i))
        if i == 0:
            assert manager.c.layout.info()["layout_info"] == [(0, 0, 800, 600)]
        elif i == 4:
            # the rest test col order
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 400, 200),
                (0, 200, 400, 200),
                (0, 400, 400, 200),
                (400, 0, 400, 300),
                (400, 300, 400, 300),
            ]
        elif i == 5:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 400, 200),
                (0, 200, 400, 200),
                (0, 400, 400, 200),
                (400, 0, 400, 200),
                (400, 200, 400, 200),
                (400, 400, 400, 200),
            ]

        elif i == 9:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 266, 150),
                (0, 150, 266, 150),
                (0, 300, 266, 150),
                (0, 450, 266, 150),
                (266, 0, 266, 150),
                (266, 150, 266, 150),
                (266, 300, 266, 150),
                (266, 450, 266, 150),
                (532, 0, 266, 300),
                (532, 300, 266, 300),
            ]
        elif i == 10:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 266, 150),
                (0, 150, 266, 150),
                (0, 300, 266, 150),
                (0, 450, 266, 150),
                (266, 0, 266, 150),
                (266, 150, 266, 150),
                (266, 300, 266, 150),
                (266, 450, 266, 150),
                (532, 0, 266, 200),
                (532, 200, 266, 200),
                (532, 400, 266, 200),
            ]
        elif i == 11:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 266, 150),
                (0, 150, 266, 150),
                (0, 300, 266, 150),
                (0, 450, 266, 150),
                (266, 0, 266, 150),
                (266, 150, 266, 150),
                (266, 300, 266, 150),
                (266, 450, 266, 150),
                (532, 0, 266, 150),
                (532, 150, 266, 150),
                (532, 300, 266, 150),
                (532, 450, 266, 150),
            ]


@ratiotile_config
def test_ratiotile_basic(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")
    sleep(0.1)
    assert manager.c.window.info()["width"] == 264
    assert manager.c.window.info()["height"] == 598
    assert manager.c.window.info()["x"] == 0
    assert manager.c.window.info()["y"] == 0
    assert manager.c.window.info()["name"] == "three"

    manager.c.group.next_window()
    assert manager.c.window.info()["width"] == 264
    assert manager.c.window.info()["height"] == 598
    assert manager.c.window.info()["x"] == 266
    assert manager.c.window.info()["y"] == 0
    assert manager.c.window.info()["name"] == "two"

    manager.c.group.next_window()
    assert manager.c.window.info()["width"] == 266
    assert manager.c.window.info()["height"] == 598
    assert manager.c.window.info()["x"] == 532
    assert manager.c.window.info()["y"] == 0
    assert manager.c.window.info()["name"] == "one"


@ratiotile_config
def test_ratiotile_window_focus_cycle(manager):
    # setup 3 tiled and two floating clients
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("float1")
    manager.c.window.toggle_floating()
    manager.test_window("float2")
    manager.c.window.toggle_floating()
    manager.test_window("three")

    # test preconditions, RatioTile adds clients to head
    assert manager.c.layout.info()["clients"] == ["three", "two", "one"]
    # last added window has focus
    assert_focused(manager, "three")

    # assert window focus cycle, according to order in layout
    assert_focus_path(manager, "two", "one", "float1", "float2", "three")


@ratiotile_config
def test_ratiotile_alternative_calculation(manager):
    manager.c.next_layout()
    manager.c.next_layout()

    for i in range(12):
        manager.test_window(str(i))
        if i == 0:
            assert manager.c.layout.info()["layout_info"] == [(0, 0, 800, 600)]
        elif i == 4:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 400, 200),
                (0, 200, 400, 200),
                (0, 400, 400, 200),
                (400, 0, 400, 300),
                (400, 300, 400, 300),
            ]
        elif i == 5:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 400, 200),
                (0, 200, 400, 200),
                (0, 400, 400, 200),
                (400, 0, 400, 200),
                (400, 200, 400, 200),
                (400, 400, 400, 200),
            ]
        elif i == 9:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 266, 150),
                (0, 150, 266, 150),
                (0, 300, 266, 150),
                (0, 450, 266, 150),
                (266, 0, 267, 200),
                (266, 200, 267, 200),
                (266, 400, 267, 200),
                (533, 0, 267, 200),
                (533, 200, 267, 200),
                (533, 400, 267, 200),
            ]
        elif i == 10:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 266, 150),
                (0, 150, 266, 150),
                (0, 300, 266, 150),
                (0, 450, 266, 150),
                (266, 0, 267, 150),
                (266, 150, 267, 150),
                (266, 300, 267, 150),
                (266, 450, 267, 150),
                (533, 0, 267, 200),
                (533, 200, 267, 200),
                (533, 400, 267, 200),
            ]
        elif i == 11:
            assert manager.c.layout.info()["layout_info"] == [
                (0, 0, 266, 150),
                (0, 150, 266, 150),
                (0, 300, 266, 150),
                (0, 450, 266, 150),
                (266, 0, 267, 150),
                (266, 150, 267, 150),
                (266, 300, 267, 150),
                (266, 450, 267, 150),
                (533, 0, 267, 150),
                (533, 150, 267, 150),
                (533, 300, 267, 150),
                (533, 450, 267, 150),
            ]


@ratiotile_config
def test_shuffling(manager):
    def clients():
        return manager.c.layout.info()["clients"]

    for i in range(3):
        manager.test_window(str(i))

    assert clients() == ["2", "1", "0"]
    manager.c.layout.shuffle_up()
    assert clients() == ["0", "2", "1"]
    manager.c.layout.shuffle_up()
    assert clients() == ["1", "0", "2"]
    manager.c.layout.shuffle_down()
    assert clients() == ["0", "2", "1"]
    manager.c.layout.shuffle_down()
    assert clients() == ["2", "1", "0"]


@ratiotile_config
def test_resizing(manager):
    def sizes():
        return manager.c.layout.info()["layout_info"]

    for i in range(5):
        manager.test_window(str(i))

    assert sizes() == [
        (0, 0, 160, 600),
        (160, 0, 160, 600),
        (320, 0, 160, 600),
        (480, 0, 160, 600),
        (640, 0, 160, 600),
    ]

    manager.c.layout.increase_ratio()
    assert sizes() == [
        (0, 0, 266, 300),
        (266, 0, 266, 300),
        (532, 0, 268, 300),
        (0, 300, 400, 300),
        (400, 300, 400, 300),
    ]

    manager.c.layout.decrease_ratio()
    assert sizes() == [
        (0, 0, 160, 600),
        (160, 0, 160, 600),
        (320, 0, 160, 600),
        (480, 0, 160, 600),
        (640, 0, 160, 600),
    ]



"""
CFFI definitions for implementing wlr_buffer with cairo surfaces
"""

SOURCE = """
#include <drm_fourcc.h>
#include <wlr/interfaces/wlr_buffer.h>

struct cairo_buffer {
    struct wlr_buffer base;
    void *data;
    size_t stride;
};

static void handle_destroy(struct wlr_buffer *wlr_buffer) {
    struct cairo_buffer *buffer = wl_container_of(wlr_buffer, buffer, base);
    free(buffer);
}

static bool handle_begin_data_ptr_access(struct wlr_buffer *wlr_buffer,
        uint32_t flags, void **data, uint32_t *format, size_t *stride) {
    struct cairo_buffer *buffer = wl_container_of(wlr_buffer, buffer, base);
    *data = buffer->data;
    *stride = buffer->stride;
    *format = DRM_FORMAT_ARGB8888;
    return true;
}

static void handle_end_data_ptr_access(struct wlr_buffer *wlr_buffer) {
    // This space is intentionally left blank
}

static const struct wlr_buffer_impl cairo_buffer_impl = {
    .destroy = handle_destroy,
    .begin_data_ptr_access = handle_begin_data_ptr_access,
    .end_data_ptr_access = handle_end_data_ptr_access,
};

struct wlr_buffer *cairo_buffer_create(int width, int height, size_t stride, void *data) {
    struct cairo_buffer *cairo_buffer = calloc(1, sizeof(struct cairo_buffer));
    if (cairo_buffer == NULL) {
        return NULL;
    }

    wlr_buffer_init(&cairo_buffer->base, &cairo_buffer_impl, width, height);
    cairo_buffer->data = data;
    cairo_buffer->stride = stride;
    return &cairo_buffer->base;
}
"""

CDEF = """
struct wlr_buffer *cairo_buffer_create(int width, int height, size_t stride, void *data);
"""



# Copyright (c) 2022 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import pytest

import libqtile.config
from libqtile import layout
from libqtile.config import Match
from libqtile.confreader import Config
from test.layouts.layout_utils import assert_dimensions


@pytest.fixture(scope="function")
def ss_manager(manager_nospawn, request):
    class ScreenSplitConfig(Config):
        auto_fullscreen = True
        groups = [libqtile.config.Group("a")]
        layouts = [layout.ScreenSplit(**getattr(request, "param", dict()))]
        floating_layout = libqtile.resources.default_config.floating_layout
        keys = []
        mouse = []
        screens = []
        follow_mouse_focus = False

    manager_nospawn.start(ScreenSplitConfig)

    yield manager_nospawn


def ss_config(**kwargs):
    return pytest.mark.parametrize("ss_manager", [kwargs], indirect=True)


@ss_config()
def test_screensplit(ss_manager):
    # Max layout is default, occupies top half of screen
    assert ss_manager.c.layout.info()["current_layout"] == "max"
    ss_manager.test_window("one")
    assert_dimensions(ss_manager, 0, 0, 800, 300)
    ss_manager.test_window("two")
    assert_dimensions(ss_manager, 0, 0, 800, 300)
    assert ss_manager.c.layout.info()["current_clients"] == ["one", "two"]

    ss_manager.c.layout.next_split()
    assert ss_manager.c.layout.info()["current_layout"] == "columns"
    assert ss_manager.c.layout.info()["current_clients"] == []

    # Columns layout has no border on single...
    ss_manager.test_window("three")
    assert_dimensions(ss_manager, 0, 300, 800, 300)
    # ... but a border of 2 when multiple windows
    ss_manager.test_window("four")
    assert_dimensions(ss_manager, 400, 300, 396, 296)
    assert ss_manager.c.layout.info()["current_clients"] == ["three", "four"]

    ss_manager.c.layout.next_split()
    assert ss_manager.c.layout.info()["current_layout"] == "max"
    assert ss_manager.c.layout.info()["current_clients"] == ["one", "two"]


@ss_config()
def test_commands_passthrough(ss_manager):
    assert ss_manager.c.layout.info()["current_layout"] == "max"
    assert "grow_left" not in ss_manager.c.layout.commands()

    ss_manager.c.layout.next_split()
    assert ss_manager.c.layout.info()["current_layout"] == "columns"

    ss_manager.test_window("one")
    assert_dimensions(ss_manager, 0, 300, 800, 300)
    ss_manager.test_window("two")
    assert_dimensions(ss_manager, 400, 300, 396, 296)

    assert "grow_left" in ss_manager.c.layout.commands()
    # Grow window by 40 pixels
    ss_manager.c.layout.grow_left()
    assert_dimensions(ss_manager, 360, 300, 436, 296)


@ss_config()
def test_move_window_to_split(ss_manager):
    assert ss_manager.c.layout.info()["current_layout"] == "max"
    ss_manager.test_window("one")
    assert_dimensions(ss_manager, 0, 0, 800, 300)

    ss_manager.c.layout.move_window_to_next_split()
    assert ss_manager.c.layout.info()["current_layout"] == "columns"
    assert_dimensions(ss_manager, 0, 300, 800, 300)

    ss_manager.c.layout.move_window_to_previous_split()
    assert ss_manager.c.layout.info()["current_layout"] == "max"
    assert_dimensions(ss_manager, 0, 0, 800, 300)


@ss_config(
    splits=[
        {
            "name": "no_match",
            "rect": (0, 0, 1, 0.5),
            "layout": layout.Max(),
        },
        {
            "name": "match",
            "rect": (0, 0.5, 1, 0.5),
            "layout": layout.Spiral(),
            "matches": [Match(title="test")],
        },
    ]
)
def test_match_window(ss_manager):
    assert ss_manager.c.layout.info()["current_layout"] == "max"
    ss_manager.test_window("one")
    assert ss_manager.c.layout.info()["current_layout"] == "max"

    ss_manager.test_window("test")
    assert ss_manager.c.layout.info()["current_layout"] == "spiral"


def test_invalid_splits():
    # Test 1: Missing required keys
    with pytest.raises(ValueError) as e:
        layout.ScreenSplit(splits=[{"rect": (0, 0, 1, 1)}])

    assert str(e.value) == "Splits must define 'name', 'rect' and 'layout'."

    # Test 2: rect is not list/tuple
    with pytest.raises(ValueError) as e:
        layout.ScreenSplit(
            splits=[{"name": "test", "rect": "0, 0, 1, 1", "layout": layout.Max()}]
        )

    assert str(e.value) == "Split rect should be a list/tuple."

    # Test 3: Wrong number of items in rect
    with pytest.raises(ValueError) as e:
        layout.ScreenSplit(splits=[{"name": "test", "rect": (0, 0, 1), "layout": layout.Max()}])

    assert str(e.value) == "Split rect should have 4 float/int members."

    # Test 4: Not all rect items are numbers
    with pytest.raises(ValueError) as e:
        layout.ScreenSplit(
            splits=[{"name": "test", "rect": (0, 0, 1, "1"), "layout": layout.Max()}]
        )

    assert str(e.value) == "Split rect should have 4 float/int members."

    # Test 5: Nested ScreenSplit
    with pytest.raises(ValueError) as e:
        layout.ScreenSplit(
            splits=[{"name": "test", "rect": (0, 0, 1, 1), "layout": layout.ScreenSplit()}]
        )

    assert str(e.value) == "ScreenSplit layouts cannot be nested."

    # Test 6: Matches has invalid object
    with pytest.raises(ValueError) as e:
        layout.ScreenSplit(
            splits=[
                {"name": "test", "rect": (0, 0, 1, 1), "layout": layout.Max(), "matches": [True]}
            ]
        )

    assert str(e.value) == "Invalid object in 'matches'."

    # Test 7: Single match
    with pytest.raises(ValueError) as e:
        layout.ScreenSplit(
            splits=[
                {
                    "name": "test",
                    "rect": (0, 0, 1, 1),
                    "layout": layout.Max(),
                    "matches": Match(wm_class="test"),
                }
            ]
        )

    assert str(e.value) == "'matches' must be a list of 'Match' objects."

    # Test 8: Test valid config - no matches
    s_split = layout.ScreenSplit(
        splits=[{"name": "test", "rect": (0, 0, 1, 1), "layout": layout.Max()}]
    )
    assert s_split

    # Test 9: Test valid config - matches
    s_split = layout.ScreenSplit(
        splits=[
            {
                "name": "test",
                "rect": (0, 0, 1, 1),
                "layout": layout.Max(),
                "matches": [Match(wm_class="test")],
            }
        ]
    )
    assert s_split



# Copyright (c) 2022 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
import shutil

import pytest

from libqtile.widget import Volume
from test.widgets.conftest import DATA_DIR

ICON = os.path.join(DATA_DIR, "svg", "audio-volume-muted.svg")
TEMP_DIR = os.path.join(DATA_DIR, "ss_temp")


@pytest.fixture
def widget():
    os.mkdir(TEMP_DIR)
    for i in (
        "audio-volume-high.svg",
        "audio-volume-low.svg",
        "audio-volume-medium.svg",
        "audio-volume-muted.svg",
    ):
        shutil.copy(ICON, os.path.join(TEMP_DIR, i))

    yield Volume

    shutil.rmtree(TEMP_DIR)


@pytest.mark.parametrize(
    "screenshot_manager",
    [{"theme_path": TEMP_DIR}, {"emoji": True}, {"fmt": "Vol: {}"}],
    indirect=True,
)
def ss_volume(screenshot_manager):
    widget = screenshot_manager.c.widget["volume"]
    widget.eval("self.volume=-1")
    widget.eval("self._update_drawer()")
    widget.eval("self.bar.draw()")
    screenshot_manager.take_screenshot()



# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2014 Sean Vig
# Copyright (c) 2014 Tycho Andersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from pathlib import Path

import pytest

from libqtile import config, confreader, utils

configs_dir = Path(__file__).resolve().parent / "configs"


def load_config(name):
    f = confreader.Config(configs_dir / name)
    f.load()
    return f


def test_validate():
    # bad key
    f = load_config("basic.py")
    f.keys[0].key = "nonexistent"
    with pytest.raises(confreader.ConfigError):
        f.validate()

    # bad modifier
    f = load_config("basic.py")
    f.keys[0].modifiers = ["nonexistent"]
    with pytest.raises(confreader.ConfigError):
        f.validate()


def test_basic():
    f = load_config("basic.py")
    assert f.keys


def test_syntaxerr():
    with pytest.raises(SyntaxError):
        load_config("syntaxerr.py")


def test_falls_back():
    f = load_config("basic.py")
    # We just care that it has a default, we don't actually care what the
    # default is; don't assert anything at all about the default in case
    # someone changes it down the road.
    assert hasattr(f, "follow_mouse_focus")


def cmd(x):
    return None


def test_ezkey():
    key = config.EzKey("M-A-S-a", cmd, cmd)
    modkey, altkey = (config.EzConfig.modifier_keys[i] for i in "MA")
    assert key.modifiers == [modkey, altkey, "shift"]
    assert key.key == "a"
    assert key.commands == (cmd, cmd)

    key = config.EzKey("M-<Tab>", cmd)
    assert key.modifiers == [modkey]
    assert key.key == "Tab"
    assert key.commands == (cmd,)

    with pytest.raises(utils.QtileError):
        config.EzKey("M--", cmd)

    with pytest.raises(utils.QtileError):
        config.EzKey("Z-Z-z", cmd)

    with pytest.raises(utils.QtileError):
        config.EzKey("asdf", cmd)

    with pytest.raises(utils.QtileError):
        config.EzKey("M-a-A", cmd)


def test_ezclick_ezdrag():
    btn = config.EzClick("M-1", cmd)
    assert btn.button == "Button1"
    assert btn.modifiers == [config.EzClick.modifier_keys["M"]]

    btn = config.EzDrag("A-2", cmd)
    assert btn.button == "Button2"
    assert btn.modifiers == [config.EzClick.modifier_keys["A"]]



# Copyright (c) 2022 m-col
# Copyright (c) 2023 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

import cairocffi
from wlroots.util.region import PixmanRegion32

from libqtile.backend.base import drawer

if TYPE_CHECKING:
    from libqtile.backend.wayland.window import Internal
    from libqtile.core.manager import Qtile


class Drawer(drawer.Drawer):
    """
    A helper class for drawing and text layout.

    1. We stage drawing operations locally in memory using a cairo RecordingSurface.
    2. Then apply these operations to the windows's underlying ImageSurface.
    """

    def __init__(self, qtile: Qtile, win: Internal, width: int, height: int):
        drawer.Drawer.__init__(self, qtile, win, width, height)

    def _draw(
        self,
        offsetx: int = 0,
        offsety: int = 0,
        width: int | None = None,
        height: int | None = None,
        src_x: int = 0,
        src_y: int = 0,
    ) -> None:
        if offsetx > self._win.width:
            return

        # We need to set the current draw area so we can compare to the previous one
        self.current_rect = (offsetx, offsety, width, height)
        # rect_changed = current_rect != self.previous_rect

        if not self.needs_update:
            return

        # Keep track of latest rect covered by this drawwer
        self.previous_rect = self.current_rect

        # Make sure geometry doesn't extend beyond texture
        if width is None:
            width = self.width
        if width > self._win.width - offsetx:
            width = self._win.width - offsetx
        if height is None:
            height = self.height
        if height > self._win.height - offsety:
            height = self._win.height - offsety

        # Paint recorded operations to our window's underlying ImageSurface
        with cairocffi.Context(self._win.surface) as context:
            context.set_operator(cairocffi.OPERATOR_SOURCE)
            # Adjust the source surface position by src_x and src_y e.g. if we want
            # to render part of the surface in a different position
            context.set_source_surface(self.surface, offsetx - src_x, offsety - src_y)
            context.rectangle(offsetx, offsety, width, height)
            context.fill()

        damage = PixmanRegion32()
        damage.init_rect(offsetx, offsety, width, height)
        # TODO: do we really need to `set_buffer` here? would be good to just set damage
        self._win._scene_buffer.set_buffer_with_damage(self._win.wlr_buffer, damage)
        damage.fini()
